from abc import ABC, abstractmethod
import asyncio
import contextlib
from datetime import datetime, timedelta
import json
import os
from pathlib import Path
import time
import traceback
from typing import Any, cast, Dict, List, Optional, Tuple, Type, Union
from urllib.parse import urlparse

from aiohttp import ClientResponseError, ClientSession, WSMsgType
from aiohttp.web import HTTPNotFound
import click
from pydantic import BaseModel, Field, root_validator
import requests
from requests.adapters import HTTPAdapter
import tabulate
from urllib3 import Retry
import yaml

from anyscale.cli_logger import LogsLogger
from anyscale.client.openapi_client import (
    ComputeTemplate,
    DecoratedProductionJob,
    HaJobStates,
    JobsLogsQueryInfo,
    ProductionJob,
    ProductionJobConfig,
)
from anyscale.client.openapi_client.models.create_internal_production_job import (
    CreateInternalProductionJob,
)
from anyscale.client.openapi_client.models.log_filter import LogFilter
from anyscale.cluster_compute import (
    get_cluster_compute_from_name,
    get_default_cluster_compute,
    register_compute_template,
)
from anyscale.cluster_env import (
    get_build_from_cluster_env_identifier,
    get_default_cluster_env_build,
    validate_successful_build,
)
from anyscale.controllers.base_controller import BaseController
from anyscale.controllers.logs_controller import LogsController
from anyscale.project import get_proj_id_from_name, infer_project_id
from anyscale.shared_anyscale_utils.utils.byod import BYODInfo
from anyscale.util import (
    extract_versions_from_image_name,
    get_endpoint,
    is_anyscale_workspace,
    poll,
    populate_dict_with_workspace_config_if_exists,
    PROJECT_NAME_ENV_VAR,
    validate_job_config_dict,
)
from anyscale.utils.runtime_env import override_runtime_env_for_local_working_dir


log = LogsLogger()

terminal_state = {
    HaJobStates.SUCCESS,
    HaJobStates.TERMINATED,
    HaJobStates.BROKEN,
    HaJobStates.OUT_OF_RETRIES,
}
pending_states = {
    HaJobStates.PENDING,
    HaJobStates.AWAITING_CLUSTER_START,
    HaJobStates.ERRORED,
    HaJobStates.RESTARTING,
}


class MiniJobRun(BaseModel):
    last_job_run_id: str
    job_state: str
    error: Optional[str]


def _validate_conda_option(conda_option: Union[str, Dict]) -> Union[str, Dict]:
    """Parses and validates a user-provided 'conda' option.

    Can be one of three cases:
        1) A str that's the name of a pre-installed conda environment.
        2) A string pointing to a local conda environment YAML. In this case,
           the file contents will be read into a dict.
        3) A dict that defines a conda environment. This is passed through.
    """
    result = None
    if isinstance(conda_option, str):
        yaml_file = Path(conda_option)
        if yaml_file.suffix in (".yaml", ".yml"):
            if not yaml_file.is_file():
                raise click.ClickException(f"Can't find conda YAML file {yaml_file}.")
            try:
                result = yaml.safe_load(yaml_file.read_text())
            except Exception as e:  # noqa: BLE001
                raise click.ClickException(
                    f"Failed to read conda file {yaml_file}: {e}."
                )
        else:
            # Assume it's a pre-existing conda environment name.
            result = conda_option
    elif isinstance(conda_option, dict):
        result = conda_option

    return result


def _validate_pip_option(pip_option: Union[str, List[str]]) -> Optional[List[str]]:
    """Parses and validates a user-provided 'pip' option.

    Can be one of two cases:
        1) A List[str] describing the requirements. This is passed through.
        2) A string pointing to a local requirements file. In this case, the
           file contents will be read split into a list.
    """
    result = None
    if isinstance(pip_option, str):
        # We have been given a path to a requirements.txt file.
        pip_file = Path(pip_option)
        if not pip_file.is_file():
            raise click.ClickException(f"{pip_file} is not a valid file.")
        result = pip_file.read_text().strip().split("\n")
    elif isinstance(pip_option, list) and all(
        isinstance(dep, str) for dep in pip_option
    ):
        if len(pip_option) == 0:
            result = None
        else:
            result = pip_option

    return result


def _validate_py_modules(py_modules_option: List[str]) -> List[str]:
    for entry in py_modules_option:
        if "://" not in entry:
            raise click.ClickException(
                "Only remote URIs are currently supported for py_modules in the job "
                "config (not local directories). Please see "
                "https://docs.ray.io/en/master/handling-dependencies.html#remote-uris for supported options."
            )

    return py_modules_option


def _working_dir_is_remote_uri(working_dir: str) -> bool:
    return "://" in working_dir


def _validate_working_dir(working_dir_option: str) -> str:
    """If working_dir is a local directory, check that it exists."""
    # We have been given a path to a local directory.
    if (
        not _working_dir_is_remote_uri(working_dir_option)
        and not Path(working_dir_option).is_dir()
    ):
        raise click.ClickException(
            f"working_dir {working_dir_option} is not a valid local directory or remote URI."
        )

    return working_dir_option


def _validate_working_dir_and_upload_path(
    working_dir: Optional[str], upload_path: Optional[str]
):
    """Check that the combination of working_dir and upload_path is valid.

    The rule is that working_dir is a local directory if and only if upload_path is specified.
    """
    if upload_path is not None:
        if working_dir is None or _working_dir_is_remote_uri(working_dir):
            raise click.ClickException(
                f"`upload_path` was specified, but `working_dir` is not a local directory.  Received `upload_path`: {upload_path} and `working_dir`: {working_dir}."
            )
    else:
        if working_dir is not None and not _working_dir_is_remote_uri(working_dir):
            raise click.ClickException(
                f"{working_dir} is a local directory, but `upload_path` was not specified in the runtime_env. Please specify a remote URI in `upload_path`, e.g. `s3://my_bucket/my_dir`."
            )


def _validate_upload_path(upload_path: str) -> str:
    """Check that the upload path is a valid S3 or GS remote URI."""
    try:
        parsed_upload_path = urlparse(upload_path)
    except Exception as e:  # noqa: BLE001
        raise click.ClickException(
            f"Failed to parse `upload_path` {upload_path} as a URI (e.g. 's3://my_bucket/my_dir'): {e}"
        )
    if parsed_upload_path.scheme not in ["s3", "gs"]:
        raise click.ClickException(
            f"Only Amazon S3 (e.g. 's3://bucket', 's3://bucket/path') and Google Storage URIs (e.g. 'gs://bucket', 'gs://bucket/path') are supported. Received {upload_path}."
        )
    return upload_path


def _validate_env_vars(env_vars: Dict[str, str]) -> Dict[str, str]:
    for key, value in env_vars.items():
        if not isinstance(key, str):
            raise click.ClickException(
                f"env_vars key {key} is not a string. Please check the formatting."
            )
        if not isinstance(value, str):
            raise click.ClickException(
                f"env_vars value {value} is not a string. Please check the formatting."
            )

    return env_vars


class BaseHAJobConfig(BaseModel):
    """
    Base job configuration for both Production Jobs and Services.
    """

    name: Optional[str] = Field(
        None,
        description="Name of job to be submitted. Default will be used if not provided.",
    )
    description: Optional[str] = Field(
        None,
        description="Description of job to be submitted. Default will be used if not provided.",
    )
    runtime_env: Optional[Dict[str, Any]] = Field(
        None,
        description="A ray runtime env json. Your entrypoint will be run in the environment specified by this runtime env.",
    )
    build_id: Optional[str] = Field(
        None,
        description="The id of the cluster env build. This id will determine the docker image your job is run on.",
    )
    cluster_env: Optional[str] = Field(
        None,
        description="The name of the cluster environment and build revision in format `my_cluster_env:1`.",
    )
    docker: Optional[str] = Field(None, description="Custom docker image name.")
    python_version: Optional[str] = Field(
        None, description="Python version for the custom docker image."
    )
    ray_version: Optional[str] = Field(
        None, description="Ray version for the custom docker image."
    )
    compute_config_id: Optional[str] = Field(
        None,
        description="The id of the compute configuration that you want to use. This id will specify the resources required for your job",
    )
    project_id: Optional[str] = Field(
        None,
        description="The id of the project you want to use. If not specified, and no project is inferred from the directory, no project will be used.",
    )
    workspace_id: Optional[str] = Field(
        None, description="The id of the workspace that this job is submitted from.",
    )
    project: Optional[str] = Field(
        None,
        description="The name of the project you want to use. If not specified, and no project is inferred from the directory, no project will be used.",
    )
    compute_config: Optional[Union[str, Dict[str, Any]]] = Field(
        None,
        description="The name of the compute configuration that you want to use. This will specify the resources required for your job."
        "This field also accepts a one-off config as a dictionary.",
    )
    cloud: Optional[str] = Field(
        None,
        description="The cloud name to specify a default compute configuration with. This will specify the resources required for your job",
    )
    max_retries: Optional[int] = Field(
        5,
        description="The number of retries this job will attempt on failure. Set to None to set infinite retries",
    )

    @root_validator
    def fill_project_id(cls: Any, values: Any) -> Any:  # noqa: N805
        project_id, project_name = (
            values.get("project_id"),
            values.get("project"),
        )
        if project_id and project_name:
            raise click.ClickException(
                "Only one of `project_id` or `project` can be provided in the job config file. "
            )
        project_name_env_var = os.environ.get(PROJECT_NAME_ENV_VAR)
        if project_name_env_var:
            # Get project from environment variable regardless of if is provided in config
            values["project_id"] = get_proj_id_from_name(project_name_env_var)
        elif project_name:
            values["project_id"] = get_proj_id_from_name(project_name)

        return values

    @root_validator
    def fill_cluster_env_from_custom_docker(cls: Any, values: Any) -> Any:  # noqa: N805
        docker, python_version, ray_version, cluster_env, build_id = (
            values.get("docker"),
            values.get("python_version"),
            values.get("ray_version"),
            values.get("cluster_env"),
            values.get("build_id"),
        )
        if docker is not None:
            if cluster_env is not None:
                raise click.ClickException(
                    "`cluster_env` and `docker` cannot both be specified. Please only provide one"
                    "of these in the job config file."
                )
            if build_id is not None:
                raise click.ClickException(
                    "`build_id` and `docker` cannot both be specified. Please only provide one"
                    "of these in the job config file."
                )
            if python_version is None and ray_version is None:
                python_version, ray_version = extract_versions_from_image_name(docker)

            if python_version is None:
                raise click.ClickException(
                    "`python_version` should be specified when `docker` is used."
                )
            if ray_version is None:
                raise click.ClickException(
                    "`ray_version` should be specified when `docker` is used."
                )
            values["cluster_env"] = BYODInfo(
                docker_image_name=docker,
                python_version=python_version,
                ray_version=ray_version,
            ).encode()
        return values

    @root_validator
    def fill_build_id(cls: Any, values: Any) -> Any:  # noqa: N805
        build_id, cluster_env = (
            values.get("build_id"),
            values.get("cluster_env"),
        )
        if cluster_env and build_id:
            raise click.ClickException(
                "Only one of `cluster_env` or `build_id` can be provided in the job config file. "
            )
        if cluster_env:
            build_id = get_build_from_cluster_env_identifier(cluster_env).id
            values["build_id"] = build_id
        elif not build_id:
            log.info(
                "No cluster environment provided, setting default based on local Python and Ray versions."
            )
            build_id = get_default_cluster_env_build().id
            values["build_id"] = build_id
        validate_successful_build(values["build_id"])
        return values

    @root_validator
    def fill_compute_config_id(cls: Any, values: Any) -> Any:  # noqa: N805
        compute_config_id, compute_config, cloud = (
            values.get("compute_config_id"),
            values.get("compute_config"),
            values.get("cloud"),
        )
        if bool(compute_config_id) + bool(compute_config) + bool(cloud) > 1:
            raise click.ClickException(
                "Only one of `compute_config_id`, `compute_config`, or `cloud` can be provided in the job config file."
            )
        if compute_config and isinstance(compute_config, str):
            compute_config_id = get_cluster_compute_from_name(compute_config).id
        elif compute_config and isinstance(compute_config, dict):
            compute_config_id = register_compute_template(compute_config).id
        elif cloud:
            # Get default cluster compute for the specified cloud.
            compute_config_id = get_default_cluster_compute(cloud, None).id
            log.info(
                f"Using default compute config for specified cloud {cloud}: {compute_config_id}."
            )
        elif not compute_config_id:
            # Get default cluster compute for the default cloud.
            compute_config_id = get_default_cluster_compute(None, None).id
            log.info(
                f"No cloud or compute config specified, using the default: {compute_config_id}."
            )
        values["compute_config_id"] = compute_config_id

        return values

    @root_validator
    def validate_runtime_env(cls: Any, values: Any) -> Any:  # noqa: N805, PLR0912
        runtime_env = values.get("runtime_env")
        if runtime_env is not None:
            if "conda" in runtime_env:
                conda_option = runtime_env["conda"]
                if not isinstance(conda_option, (str, dict)):
                    raise click.ClickException(
                        f"runtime_env['conda'] must be str or dict, got type({conda_option})."
                    )
                runtime_env["conda"] = _validate_conda_option(conda_option)
            if "pip" in runtime_env:
                pip_option = runtime_env["pip"]
                if not isinstance(pip_option, (str, list)):
                    raise click.ClickException(
                        f"runtime_env['pip'] must be str or list, got type({pip_option})."
                    )
                runtime_env["pip"] = _validate_pip_option(runtime_env["pip"])
            if "py_modules" in runtime_env:
                py_modules_option = runtime_env["py_modules"]
                if not isinstance(py_modules_option, list):
                    raise click.ClickException(
                        f"runtime_env['py_modules'] must be list, got type({py_modules_option})."
                    )
                runtime_env["py_modules"] = _validate_py_modules(py_modules_option)
            if "upload_path" in runtime_env:
                upload_path_option = runtime_env["upload_path"]
                if not isinstance(upload_path_option, str):
                    raise click.ClickException(
                        f"runtime_env['upload_path'] must be str, got type({upload_path_option})."
                    )
                runtime_env["upload_path"] = _validate_upload_path(upload_path_option)
            if "working_dir" in runtime_env:
                working_dir_option = runtime_env["working_dir"]
                if not isinstance(working_dir_option, str):
                    raise click.ClickException(
                        f"runtime_env['working_dir'] must be str, got type({working_dir_option})."
                    )
                runtime_env["working_dir"] = _validate_working_dir(working_dir_option)
            _validate_working_dir_and_upload_path(
                runtime_env.get("working_dir"), runtime_env.get("upload_path")
            )
            if "env_vars" in runtime_env:
                env_vars_option = runtime_env["env_vars"]
                if not isinstance(env_vars_option, dict):
                    raise click.ClickException(
                        f"runtime_env['env_vars'] must be dict, got type({env_vars_option})."
                    )
                runtime_env["env_vars"] = _validate_env_vars(env_vars_option)
            values["runtime_env"] = runtime_env

        return values


class JobConfig(BaseHAJobConfig):
    """
    Job Config model for CLI. Validate and preprocess so `entrypoint`, `runtime_env_config`,
    `build_id`, `compute_config_id`, `max_retries` have the correct values to call
    `/api/v2/decorated_ha_jobs/create`.
    """

    entrypoint: str = Field(
        ...,
        description="A script that will be run to start your job. This command will be run in the root directory of the specified runtime env. Eg. 'python script.py'",
    )


class LogProvider(ABC):
    @abstractmethod
    def __init__(self, dns_name: str, anyscale_token: str, query: str) -> None:
        pass

    @abstractmethod
    def query(
        self, start_timestamp_ns: int, end_timestamp_ns: int
    ) -> List[Tuple[int, str]]:
        """Query logs for a time range, return list of (timestamp_ns, log_line)."""

    @abstractmethod
    def close(self) -> None:
        """Close any connection this log provider uses"""

    def __enter__(self, *args, **kwargs):  # noqa: B027
        pass

    def __exit__(self, *args, **kwargs):  # noqa: ARG002
        self.close()


class LokiLogProvider(LogProvider):
    def __init__(self, dns_name: str, anyscale_token: str, query: str) -> None:
        # https://grafana.com/docs/loki/latest/api/#get-lokiapiv1query_range
        self.url = f"https://{dns_name}/loki/api/v1/query_range"
        self.params: Dict[str, Union[str, int]] = {
            "token": anyscale_token,
            "query": query,
            "limit": 1000,
            "start": 0,
            "end": int(time.time() * 1e9),
            "direction": "forward",
        }

    def __enter__(self, *args, **kwargs):  # noqa: ARG002
        self.session = requests.Session()
        self.session.mount(
            "https://",
            HTTPAdapter(
                max_retries=Retry(total=10, backoff_factor=0.1, allowed_methods=["GET"])
            ),
        )
        return self

    def close(self) -> None:
        self.session.close()

    def __exit__(self, *args, **kwargs):  # noqa: ARG002
        self.close()

    def query(
        self, start_timestamp_ns: int, end_timestamp_ns: int
    ) -> List[Tuple[int, str]]:
        if start_timestamp_ns == 0:
            # 10^9 nanoseconds in a second
            start_timestamp_ns = int(
                (datetime.now() - timedelta(days=7)).timestamp() * (10 ** 9)
            )
        self.params["start"] = start_timestamp_ns
        self.params["end"] = end_timestamp_ns

        resp = self.session.get(self.url, params=self.params)
        resp.raise_for_status()

        resp_data = resp.json()
        if resp_data["status"] != "success":
            raise click.ClickException("Querying Anyscale log endpoint failed.")
        if resp_data["data"]["resultType"] != "streams":
            raise click.ClickException("Invalid metrics type.")

        lines = []
        for log_chunk in resp_data["data"]["result"]:
            # TODO(simon): potentially annotate the log with the log_chunk["stream"] metadata
            for timestamp, line in log_chunk["values"]:
                lines.append((int(timestamp), line))

        return lines


class JobController(BaseController):
    def __init__(
        self, log: Optional[LogsLogger] = None, initialize_auth_api_client: bool = True
    ):
        if log is None:
            log = LogsLogger()

        super().__init__(initialize_auth_api_client=initialize_auth_api_client)
        self.log = log
        self.log.open_block("Output")
        self.logs_controller = LogsController()

    def submit(  # noqa: PLR0913
        self,
        job_config_file: str,
        name: Optional[str] = None,
        description: Optional[str] = None,
        is_entrypoint_cmd: Optional[bool] = False,
        entrypoint: Optional[List[str]] = None,
    ) -> str:
        entrypoint = entrypoint or []
        workspace_id = os.environ.get("ANYSCALE_EXPERIMENTAL_WORKSPACE_ID", None)
        if is_anyscale_workspace() and is_entrypoint_cmd:
            entrypoint = [job_config_file, *entrypoint]
            config = self.generate_config_from_entrypoint(
                entrypoint, name, description, workspace_id
            )
            id = self.submit_from_config(config)  # noqa: A001
        elif len(entrypoint) == 0:
            # Assume that job_config_file is a file and submit it.
            config = self.generate_config_from_file(
                job_config_file,
                name=name,
                description=description,
                workspace_id=workspace_id,
            )
            id = self.submit_from_config(config)  # noqa: A001
        elif len(entrypoint) != 0:
            msg = (
                "Within an Anyscale Workspace, `anyscale job submit` takes either a file, or a command. To submit a command, use `anyscale job submit -- my command`."
                if is_anyscale_workspace()
                else "`anyscale job submit` takes one argument, a YAML file configuration. Please use `anyscale job submit my_file`."
            )
            raise click.ClickException(msg)
        return id

    def generate_config_from_entrypoint(
        self,
        entrypoint: List[str],
        name: Optional[str],
        description: Optional[str],
        workspace_id: Optional[str] = None,
    ) -> JobConfig:
        config_dict = {
            "entrypoint": " ".join(entrypoint),
            "name": name,
            "description": description,
            "workspace_id": workspace_id,
        }
        config_dict = populate_dict_with_workspace_config_if_exists(
            config_dict, self.anyscale_api_client
        )
        job_config = JobConfig.parse_obj(config_dict)
        return job_config

    def generate_config_from_file(
        self,
        job_config_file: str,
        name: Optional[str],
        description: Optional[str],
        workspace_id: Optional[str] = None,
    ) -> JobConfig:
        config_dict = self._load_config_dict_from_file(job_config_file)
        config_dict["workspace_id"] = workspace_id
        validate_job_config_dict(config_dict, self.api_client)
        config_dict = populate_dict_with_workspace_config_if_exists(
            config_dict, self.anyscale_api_client
        )
        job_config = JobConfig.parse_obj(config_dict)
        if name:
            job_config.name = name

        if description:
            job_config.description = description

        return job_config

    def submit_from_config(self, job_config: JobConfig):
        # If project id is not specified, try to infer it
        project_id = infer_project_id(
            job_config.project_id, self.anyscale_api_client, self.log
        )
        job_config.runtime_env = override_runtime_env_for_local_working_dir(
            job_config.runtime_env
        )

        config_object = ProductionJobConfig(
            entrypoint=job_config.entrypoint,
            runtime_env=job_config.runtime_env,
            build_id=job_config.build_id,
            compute_config_id=job_config.compute_config_id,
            max_retries=job_config.max_retries,
        )

        job = self.api_client.create_job_api_v2_decorated_ha_jobs_create_post(
            CreateInternalProductionJob(
                name=job_config.name or f"cli-job-{datetime.now().isoformat()}",
                description=job_config.description or "Job submitted from CLI",
                project_id=project_id,
                workspace_id=job_config.workspace_id,
                config=config_object,
            )
        ).result
        self.log.info(
            f"Maximum uptime is {self._get_maximum_uptime_output(job)} for clusters launched by this job."
        )
        self.log.info(
            f"Job {job.id} has been successfully submitted. Current state of job: {job.state.current_state}."
        )
        self.log.info(
            f"Query the status of the job with `anyscale job list --job-id {job.id}`."
        )
        self.log.info(
            f"Get the logs for the job with `anyscale job logs --job-id {job.id} --follow`."
        )
        self.log.info(f'View the job in the UI at {get_endpoint(f"/jobs/{job.id}")}.')
        return job.id

    def _get_maximum_uptime_output(self, job: ProductionJob) -> str:
        compute_config: ComputeTemplate = self.api_client.get_compute_template_api_v2_compute_templates_template_id_get(
            job.config.compute_config_id
        ).result
        maximum_uptime_minutes = compute_config.config.maximum_uptime_minutes
        if maximum_uptime_minutes and maximum_uptime_minutes > 0:
            return f"set to {maximum_uptime_minutes} minutes"
        return "disabled"

    def _load_config_dict_from_file(self, job_config_file: str) -> Dict[str, Any]:
        if not os.path.exists(job_config_file):
            raise click.ClickException(f"Config file {job_config_file} not found.")

        with open(job_config_file) as f:
            config_dict = yaml.safe_load(f)
        return config_dict

    def wait(  # noqa: PLR0913
        self,
        job_name: Optional[str] = None,
        job_id: Optional[str] = None,
        is_service: bool = False,
        target_state: str = HaJobStates.SUCCESS,
        interval_secs: float = 10.0,
        timeout_secs=None,
    ):
        if target_state not in HaJobStates.allowable_values:
            raise click.ClickException(
                f"`{target_state}` is not a valid Job state. Allowed states are {HaJobStates.allowable_values}"
            )
        job_id_or_name = job_name or job_id
        job_id = cast(
            str,
            self._resolve_job_object(
                job_id, job_name, entity_type="service" if is_service else "job"
            ).id,
        )
        with self.log.spinner(f"Waiting for Job `{job_id_or_name}`...") as spinner:
            state = None
            for _ in poll(interval_secs=interval_secs, timeout_secs=timeout_secs):
                state = self._get_job_state(job_id)
                spinner.text = f"Job `{job_id_or_name}` is in state `{state}`. Waiting to reach state `{target_state}`."

                if state == target_state:
                    spinner.succeed(
                        f"Job `{job_id_or_name}` reached state `{target_state}`."
                    )
                    return job_id

                if state in terminal_state:
                    msg = f"Job `{job_id_or_name}` reached terminal state `{state}`, and will not reach `{target_state}`."
                    spinner.fail(msg)
                    raise click.ClickException(msg)

            msg = f"Timed out after waiting for {timeout_secs} seconds. The current state is {state}."
            spinner.fail(msg)
            raise click.ClickException(msg)

    def _get_job_state(self, job_id: str):
        job = self.api_client.get_job_api_v2_decorated_ha_jobs_production_job_id_get(
            job_id
        ).result
        return job.state.current_state

    def list(  # noqa: PLR0913
        self,
        include_all_users: bool,
        name: Optional[str],
        job_id: Optional[str],
        project_id: Optional[str],
        include_archived: bool,
        max_items: int,
        is_service: bool = False,
    ) -> None:
        """
        This function will either list jobs or services based on the value of `is_service`.
        Both functionalities are combined in one function because the API to list both these
        entities is currently the same.
        """
        entity_type = "services" if is_service else "jobs"
        print(f'View your {entity_type} in the UI at {get_endpoint(f"/{entity_type}")}')

        jobs_list = []
        if job_id:
            job = self.api_client.get_job_api_v2_decorated_ha_jobs_production_job_id_get(
                job_id
            ).result
            if not is_service and job.is_service:
                # `job_id` belongs to a service, but this function should list jobs.
                raise click.ClickException(
                    f"ID {job_id} belongs to a Anyscale service. Please get information about "
                    f"this service with `anyscale service list --service-id {job_id}`."
                )
            elif is_service and not job.is_service:
                # `job_id` belongs to a job, but this function should list services.
                raise click.ClickException(
                    f"ID {job_id} belongs to a Anyscale job. Please get information about "
                    f"this job with `anyscale job list --job-id {job_id}`."
                )
            output_map = {
                "Name": job.name,
                "Id": job.id,
                "Cost (dollars)": job.cost_dollars,
                "Project name": job.project.name,
                "Cluster name": job.last_job_run.cluster.name
                if job.last_job_run and job.last_job_run.cluster
                else None,
                "Current state": job.state.current_state,
                "Creator": job.creator.username,
                "Entrypoint": job.config.entrypoint
                if len(job.config.entrypoint) < 100
                else job.config.entrypoint[:100] + " ...",
            }
            if is_service:
                output_map["Access"] = job.access
                output_map["URL"] = job.url
                output_map["Token"] = job.token

            output_str = "\n".join(
                [f"\t{key}: {output_map[key]}" for key in output_map]
            )
            print(output_str)
            return
        else:
            if not include_all_users:
                creator_id = (
                    self.api_client.get_user_info_api_v2_userinfo_get().result.id
                )
            else:
                creator_id = None
            resp = self.api_client.list_decorated_jobs_api_v2_decorated_ha_jobs_get(
                project_id=project_id,
                name=name,
                creator_id=creator_id,
                type_filter="SERVICE" if is_service else "BATCH_JOB",
                archive_status="ALL" if include_archived else "NOT_ARCHIVED",
                count=10,
            )
            jobs_list.extend(resp.results)
            paging_token = resp.metadata.next_paging_token
            has_more = (paging_token is not None) and (len(jobs_list) < max_items)
            while has_more:
                resp = self.api_client.list_decorated_jobs_api_v2_decorated_ha_jobs_get(
                    project_id=project_id,
                    name=name,
                    creator_id=creator_id,
                    type_filter="SERVICE" if is_service else "BATCH_JOB",
                    archive_status="ALL" if include_archived else "NOT_ARCHIVED",
                    count=10,
                    paging_token=paging_token,
                )
                jobs_list.extend(resp.results)
                paging_token = resp.metadata.next_paging_token
                has_more = (paging_token is not None) and (len(jobs_list) < max_items)
            jobs_list = jobs_list[:max_items]

        jobs_table = [
            [
                job.name,
                job.id,
                job.cost_dollars,
                job.project.name,
                job.last_job_run.cluster.name
                if job.last_job_run and job.last_job_run.cluster
                else None,
                job.state.current_state,
                job.creator.username,
                job.config.entrypoint
                if len(job.config.entrypoint) < 50
                else job.config.entrypoint[:50] + " ...",
            ]
            for job in jobs_list
        ]

        table = tabulate.tabulate(
            jobs_table,
            headers=[
                "NAME",
                "ID",
                "COST",
                "PROJECT NAME",
                "CLUSTER NAME",
                "CURRENT STATE",
                "CREATOR",
                "ENTRYPOINT",
            ],
            tablefmt="plain",
        )
        print(f"{entity_type.capitalize()}:\n{table}")

    def archive(
        self, job_id: Optional[str], job_name: Optional[str], is_service: bool = False,
    ) -> None:
        """
        This function will either archive jobs or services based on the value of `is_service`.
        Both functionalities are combined in one function because the API to archive both these
        entities is currently the same.
        """
        entity_type = "service" if is_service else "job"
        job_resp: DecoratedProductionJob = self._resolve_job_object(
            job_id, job_name, entity_type
        )
        self.api_client.archive_job_api_v2_decorated_ha_jobs_production_job_id_archive_post(
            job_resp.id
        )
        self.log.info(
            f"{entity_type.capitalize()} {job_resp.id} is successfully archived."
        )

    def terminate(
        self, job_id: Optional[str], job_name: Optional[str], is_service: bool = False,
    ) -> None:
        """
        This function will either terminate jobs or services based on the value of `is_service`.
        Both functionalities are combined in one function because the API to terminate both these
        entities is currently the same.
        """
        entity_type = "service" if is_service else "job"
        job_resp: DecoratedProductionJob = self._resolve_job_object(
            job_id, job_name, entity_type
        )
        job = self.api_client.terminate_job_api_v2_decorated_ha_jobs_production_job_id_terminate_post(
            job_resp.id
        ).result
        self.log.info(f"{entity_type.capitalize()} {job.id} has begun terminating...")
        self.log.info(
            f" Current state of {entity_type}: {job.state.current_state}. Goal state of {entity_type}: {job.state.goal_state}"
        )
        self.log.info(
            f"Query the status of the {entity_type} with `anyscale {entity_type} list --{entity_type}-id {job.id}`."
        )

    def _resolve_job_object(
        self, job_id: Optional[str], job_name: Optional[str], entity_type: str = "job"
    ) -> DecoratedProductionJob:
        """Given job_id or job_name, retrieve decorated ha job spec"""
        if job_id is None and job_name is None:
            raise click.ClickException(
                f"Either `--id` or `--name` must be passed in for {entity_type}."
            )
        if job_id is None:
            jobs_resp = self.api_client.list_decorated_jobs_api_v2_decorated_ha_jobs_get(
                name=job_name,
                type_filter="BATCH_JOB" if entity_type == "job" else "SERVICE",
            ).results
            if len(jobs_resp) == 0:
                raise click.ClickException(
                    f"No {entity_type} found with name {job_name}. Please either pass `--id` or list the "
                    f"available {entity_type}s with `anyscale {entity_type} list`."
                )
            if len(jobs_resp) > 1:
                raise click.ClickException(
                    f"Multiple {entity_type}s found with name {job_name}. Please specify the `--id` instead."
                )
            job_id = jobs_resp[0].id
        jobs_resp = self.api_client.get_job_api_v2_decorated_ha_jobs_production_job_id_get(
            production_job_id=job_id
        ).result
        return jobs_resp

    def _get_latest_job_run(self, job_id: str) -> DecoratedProductionJob:
        job_object = self.api_client.get_job_api_v2_decorated_ha_jobs_production_job_id_get(
            production_job_id=job_id
        ).result
        return job_object

    def _get_formatted_latest_job_run(self, job_id: str) -> Optional[MiniJobRun]:
        job_object = self._get_latest_job_run(job_id)
        job_state = job_object.state.current_state
        last_job_run_id = job_object.last_job_run_id
        if job_state in terminal_state and last_job_run_id is None:
            raise click.ClickException(
                f"Can't find latest job run for {job_state} job."
            )
        if not last_job_run_id:
            return None
        return MiniJobRun(
            last_job_run_id=last_job_run_id,
            job_state=job_state,
            error=job_object.state.error,
        )

    def _wait_for_a_job_run(self, job_id: str) -> str:
        """Given job_id, wait until there is a job run on a cluster."""
        last_job_run_id = None
        with self.log.spinner("Waiting for job run...") as spinner:
            while last_job_run_id is None:
                job_object = self._get_latest_job_run(job_id)
                job_state = job_object.state.current_state
                last_job_run_id = job_object.last_job_run_id
                if job_state in terminal_state and last_job_run_id is None:
                    raise click.ClickException(
                        f"Can't find latest job run for {job_state} job."
                    )
                if last_job_run_id is None:
                    spinner.text = (
                        f"Waiting for a job run, current state is {job_state}..."
                    )
                    time.sleep(5)
        return last_job_run_id

    def _wait_for_job_running(self, job_id: str) -> None:
        with self.log.spinner("Waiting for job to retry...") as spinner:
            for _ in poll():
                job_object = self._get_latest_job_run(job_id)
                job_state = job_object.state.current_state
                if job_state in pending_states:
                    spinner.text = (
                        f"Waiting for job to retry, current state is {job_state}..."
                    )
                else:
                    return

    def stream_logs(self, job_id: str) -> None:
        job_run_id = self._wait_for_a_job_run(job_id)
        websocket_url: str = (
            self.api_client.get_job_logs_stream_api_v2_logs_job_logs_stream_job_id_get(
                job_id=job_run_id
            ).result.stream_url
        )

        with contextlib.suppress(KeyboardInterrupt):
            # run_until_complete() is used here b/c the KeyboardInterrupt exception was
            # not propogated all the way up (ie. it was being handled by internals of asyncio.run())
            asyncio.get_event_loop().run_until_complete(
                self._read_websocket(websocket_url, job_run_id)
            )

    def download_logs(self, job_id: str) -> None:
        prod_job = self._get_latest_job_run(job_id)
        if prod_job.last_job_run is not None:
            cluster_id_of_latest_job_run = prod_job.last_job_run.cluster_id
        else:
            cluster_id_of_latest_job_run = prod_job.state.cluster_id
        filter = LogFilter(  # noqa: A001
            cluster_id=cluster_id_of_latest_job_run, glob="/job-driver-"
        )
        self.logs_controller.download_logs(filter=filter, write_to_stdout=True)

    def logs(
        self,
        job_id: Optional[str] = None,
        job_name: Optional[str] = None,
        should_follow: bool = False,
    ) -> None:
        job_id = cast(str, self._resolve_job_object(job_id, job_name).id)
        if should_follow:
            try:
                self.stream_logs(job_id)
                return
            except Exception as e:  # noqa: BLE001
                log.info(f"Job logs could not be followed due to error: {e}")
        try:
            log.info("Now downloading logs from S3...")
            self.download_logs(job_id)
        except Exception as e:  # noqa: BLE001
            log.info(f"Job logs could not be downloaded due to error: {e}")
            log.info("Now trying default method of fetching logs.")
            self.loki_logs(job_id, should_follow)

    def loki_logs(
        self,
        job_id: str,
        should_follow: bool = False,
        log_provider_cls: Type[LogProvider] = LokiLogProvider,
    ) -> None:
        self._wait_for_a_job_run(job_id)

        start_timestamp_ns = 0
        max_iter = 1 if not should_follow else None
        log_provider_map: Dict[str, LogProvider] = {}

        for _ in poll(1, max_iter=max_iter):
            # Fetch the latest job run
            jr = self._get_formatted_latest_job_run(job_id)

            # This shouldn't happen because we are waiting for it above
            assert jr, "Could not find the latest job run"

            log_provider = log_provider_map.get(jr.last_job_run_id)
            if not log_provider:
                # Open a new CLI block
                # This is to ensure the logs for each job are nicely formatted
                self.log.open_block(
                    jr.last_job_run_id,
                    f"Job Run Id: {jr.last_job_run_id}",
                    auto_close=True,
                )

                # We have not seen this job run id before. We need to construct a new log provider
                log_info: JobsLogsQueryInfo = self.api_client.get_job_logs_query_info_api_v2_decorated_ha_jobs_production_job_id_logs_query_get(
                    production_job_id=job_id
                ).result

                log_provider = log_provider_cls(
                    dns_name=log_info.loki_dns_name,
                    anyscale_token=log_info.access_token,
                    query=log_info.loki_query,
                )

                # Add to the cache
                log_provider_map[jr.last_job_run_id] = log_provider

            # Make sure we only have 1 open request session at a time
            with log_provider:

                # Flush all logs in the loki endpoint
                last_received_logs_iteration = 0
                for iteration in poll(interval_secs=1):
                    # Fetch logs in the time range
                    end_timestamp_ns = int(time.time() * 1e9)
                    lines = log_provider.query(start_timestamp_ns, end_timestamp_ns)

                    # Print the results
                    for time_stamp, line in lines:
                        # Update the start timestamp in case we are following or paginated.
                        start_timestamp_ns = time_stamp + 1
                        self.log.log(line)
                    if len(lines) > 0:
                        # Record the last iteration at which we received logs
                        last_received_logs_iteration = iteration

                    # If we are following, threshold is 30 iterations (or 30 seconds)
                    threshold = 30 if should_follow else 0
                    if iteration - last_received_logs_iteration > threshold:
                        # If we haven't received logs after threshold iterations, assume existing logs have been flushed
                        break

                # We have already flushed existing logs for this job run
                # If the job is in a terminal state, exit
                if jr.job_state in terminal_state:
                    # If the state is terminal, exit
                    self.log.info(f"Job entered terminal state {jr.job_state}.")
                    if jr.error:
                        self.log.error(f"Job exited with error: {jr.error}")
                    return

                # If the job is in a pending state, wait for it to start running
                # before fetching more logs
                if jr.job_state in pending_states:
                    self._wait_for_job_running(job_id)

    # Implementation is same as
    # https://github.com/ray-project/ray/blob/6142f52ada995b7123ea7347a70613ca7b924e0c/dashboard/modules/job/sdk.py#L426
    # However Ray is only part of [all] install, so we replicate the functionality here.
    async def _read_websocket(self, websocket_url: str, job_run_id: str) -> None:
        try:
            async with ClientSession() as session:
                async with session.ws_connect(websocket_url) as ws:
                    while True:
                        msg = await ws.receive()
                        if msg.type == WSMsgType.TEXT:
                            self.log.log(msg.data.strip())
                        elif msg.type == WSMsgType.CLOSED:
                            break
                        elif msg.type == WSMsgType.ERROR:
                            pass
        except ClientResponseError as e:
            # Will return 404 if job run (Ray job) is not found
            # https://github.com/ray-project/ray/blob/6334576223dd916469fcdac185752b80e0c3e416/dashboard/modules/job/job_head.py#L485
            if e.status == HTTPNotFound.status_code:
                raise Exception(
                    f"Job run {job_run_id} not found in cluster for current running job."
                )
            raise e

    def retrieve_output(
        self, job_name: Optional[str] = None, job_id: Optional[str] = None,
    ) -> None:
        if not job_id:
            job = self._resolve_job_object(job_id, job_name)
            job_id = job.id

        try:
            output = self.api_client.find_structured_job_output_api_v2_structured_outputs_ha_job_output_ha_job_id_get(
                ha_job_id=job_id
            )
            self.log.log(json.dumps(output.result.output, indent=2))
        except Exception as e:  # noqa: BLE001
            traceback.print_exc()
            # Swallow if pretty print logic fails
            with contextlib.suppress(Exception):
                j = json.loads(e.body)  # type: ignore
                detail = j.get("error", {}).get("detail")
                self.log.error(detail)



from datetime import datetime

from anyscale.client.openapi_client.models.apply_production_service_v2_api_model import (
    ApplyProductionServiceV2APIModel,
)
from anyscale.client.openapi_client.models.apply_service_config import (
    ApplyServiceConfig,
)
from anyscale.client.openapi_client.models.create_production_service import (
    CreateProductionService,
)
from anyscale.client.openapi_client.models.production_job_config import (
    ProductionJobConfig,
)
from anyscale.client.openapi_client.models.user_service_access_types import (
    UserServiceAccessTypes as UserServiceAccessTypesAPIModel,
)
from anyscale.models.service_model import ServiceConfig, UserServiceAccessTypes


def format_service_config(service_config: ServiceConfig) -> ApplyServiceConfig:
    """
    This method formats the ServiceConfig into the API model ApplyServiceConfig.

    Please note that ServiceConfig captures Service configs for both Service v1 and v2.
    Since the CLI client can't determine whether the Service v1 or v2 should be launched
    because it depends on FFs, the CLI client will try to parse ServiceConfig into
    both the Service v1 and v2 configs. Therefore, both the Service v1 and v2 configs
    can exist at the same time.
    """
    production_job_config = ProductionJobConfig(
        entrypoint=service_config.entrypoint,
        runtime_env=service_config.runtime_env,
        build_id=service_config.build_id,
        compute_config_id=service_config.compute_config_id,
        max_retries=service_config.max_retries,
        ray_serve_config=service_config.ray_serve_config,
    )
    service_name = service_config.name or f"cli-job-{datetime.now().isoformat()}"
    access = (
        UserServiceAccessTypesAPIModel.PUBLIC
        if service_config.access == UserServiceAccessTypes.public
        else UserServiceAccessTypesAPIModel.PRIVATE
    )
    service_v1_config = CreateProductionService(
        name=service_name,
        description=service_config.description or "Service updated from CLI",
        project_id=service_config.project_id,
        config=production_job_config,
        healthcheck_url=service_config.healthcheck_url,
        access=access,
    )
    # We construct the Service v2 config model only if ray_serve_config is present
    if service_config.ray_serve_config:
        service_v2_config = ApplyProductionServiceV2APIModel(
            name=service_name,
            description=service_config.description or "Service updated from CLI",
            project_id=service_config.project_id,
            version=service_config.version,
            canary_weight=service_config.canary_weight,
            ray_serve_config=service_config.ray_serve_config,
            build_id=service_config.build_id,
            compute_config_id=service_config.compute_config_id,
        )
    else:
        service_v2_config = None
    return ApplyServiceConfig(
        service_v1_config=service_v1_config, service_v2_config=service_v2_config
    )


def format_service_config_v2(
    service_config: ServiceConfig,
) -> ApplyProductionServiceV2APIModel:
    if not service_config.ray_serve_config:
        raise RuntimeError(
            "ray_serve_config is expected for the Service configuration."
        )
    service_name = service_config.name or f"cli-job-{datetime.now().isoformat()}"

    return ApplyProductionServiceV2APIModel(
        name=service_name,
        description=service_config.description or "Service updated from CLI",
        project_id=service_config.project_id,
        version=service_config.version,
        canary_weight=service_config.canary_weight,
        ray_serve_config=service_config.ray_serve_config,
        build_id=service_config.build_id,
        compute_config_id=service_config.compute_config_id,
    )



import asyncio
import functools
import os
import socket
import sys
from typing import Any

import click
from urllib3.connection import HTTPConnection
import wrapt

from anyscale.client import openapi_client
from anyscale.client.openapi_client.api.default_api import DefaultApi
from anyscale.client.openapi_client.rest import ApiException as ApiExceptionInternal
from anyscale.sdk import anyscale_client
from anyscale.sdk.anyscale_client.api.default_api import DefaultApi as AnyscaleApi
from anyscale.sdk.anyscale_client.rest import ApiException as ApiExceptionExternal
from anyscale.shared_anyscale_utils.headers import RequestHeaders
from anyscale.version import __version__ as version


# NOTE: Some OSes don't implement all of these, so only include those that the OS supports.
_SOCKET_OPTIONS = (
    HTTPConnection.default_socket_options
    + [
        (getattr(socket, t[0]), getattr(socket, t[1]), t[2])
        for t in [
            ("SOL_SOCKET", "SO_KEEPALIVE", 1),
            # Begin sending keepalives after 15 seconds of no activity:
            ("SOL_TCP", "TCP_KEEPIDLE", 15),
            # Send keepalives every 15 seconds:
            ("SOL_TCP", "TCP_KEEPINTVL", 15),
            # Number failed keepalives before marking the connection as dead:
            ("SOL_TCP", "TCP_KEEPCNT", 6),
        ]
        if hasattr(socket, t[0]) and hasattr(socket, t[1])
    ]
    # Darwin uses TCP_KEEPALIVE[A] instead of TCP_KEEPIDLE. Unfortunatley, this was
    # only added to the socket library recently (June 2021)[B], so we include the value
    # of TCP_KEEPALIVE (16)[C] if the version of python is not super recent.
    # [A] Bug: https://bugs.python.org/issue34932
    # [B] Resolution: https://github.com/python/cpython/pull/25079
    # [C] Darwin Source: https://github.com/apple/darwin-xnu/blob/main/bsd/netinet/tcp.h
    + (
        [(6, getattr(socket, "TCP_KEEPALIVE", 16), 15)]
        if sys.platform == "darwin"
        else []
    )
)


# client is of type APIClient, which is auto-generated
def configure_tcp_keepalive(client: Any) -> None:
    assert hasattr(
        client, "rest_client"
    ), f"Incorrect object of type: {type(client)}\nThis object does not have a `rest_client` property."
    client.rest_client.pool_manager.connection_pool_kw[
        "socket_options"
    ] = _SOCKET_OPTIONS


# client is of type APIClient, which is auto-generated
def configure_open_api_client_headers(client: Any, client_name: str) -> None:
    client.set_default_header(RequestHeaders.CLIENT, client_name)
    client.set_default_header(RequestHeaders.CLIENT_VERSION, version)


class _ApiClient:
    api_client: DefaultApi = None
    anyscale_client: AnyscaleApi = None


def format_api_exception(
    e, method: str, resource_path: str, raise_structured_exception: bool = False,
) -> None:
    if os.environ.get("ANYSCALE_DEBUG") == "1" or raise_structured_exception:
        raise e
    else:
        raise click.ClickException(
            f"API Exception ({e.status}) from {method} {resource_path} \n"
            f"Reason: {e.reason}\nHTTP response body: {e.body}\n"
            f"Trace ID: {e.headers._container.get('x-trace-id', None)}"  # noqa: SLF001
        )


class ApiClientWrapperInternal(openapi_client.ApiClient):
    def __init__(self, *args, raise_structured_exception: bool = False, **kwargs):
        """
        Arguments:
            raise_structured_exception (bool): If True, API exceptions will be raised
            as structured exceptions. If this and ANYSCALE_DEBUG are False, API
            exceptions will be raised as user friendly but unstructured Click exceptions.
            This arguement allows us to determine the type of raised error in code, but
            users should ANYSCALE_DEBUG to configure this.
        """
        self.raise_structured_exception = raise_structured_exception
        super().__init__(*args, **kwargs)

    def call_api(  # noqa: PLR0913
        self,
        resource_path,
        method,
        path_params=None,
        query_params=None,
        header_params=None,
        body=None,
        post_params=None,
        files=None,
        response_type=None,
        auth_settings=None,
        async_req=None,
        _return_http_data_only=None,
        collection_formats=None,
        _preload_content=True,
        _request_timeout=None,
        _host=None,
    ):
        try:
            return openapi_client.ApiClient.call_api(
                self,
                resource_path,
                method,
                path_params,
                query_params,
                header_params,
                body,
                post_params,
                files,
                response_type,
                auth_settings,
                async_req,
                _return_http_data_only,
                collection_formats,
                _preload_content,
                _request_timeout,
                _host,
            )
        except ApiExceptionInternal as e:
            format_api_exception(
                e, method, resource_path, self.raise_structured_exception
            )


class ApiClientWrapperExternal(anyscale_client.ApiClient):
    def __init__(self, *args, raise_structured_exception: bool = False, **kwargs):
        """
        Arguments:
            raise_structured_exception (bool): If True, API exceptions will be raised
            as structured exceptions. If this and ANYSCALE_DEBUG are False, API
            exceptions will be raised as user friendly but unstructured Click exceptions.
            This arguement allows us to determine the type of raised error in code, but
            users should ANYSCALE_DEBUG to configure this.
        """
        self.raise_structured_exception = raise_structured_exception
        super().__init__(*args, **kwargs)

    def call_api(  # noqa: PLR0913
        self,
        resource_path,
        method,
        path_params=None,
        query_params=None,
        header_params=None,
        body=None,
        post_params=None,
        files=None,
        response_type=None,
        auth_settings=None,
        async_req=None,
        _return_http_data_only=None,
        collection_formats=None,
        _preload_content=True,
        _request_timeout=None,
        _host=None,
    ):
        try:
            return anyscale_client.ApiClient.call_api(
                self,
                resource_path,
                method,
                path_params,
                query_params,
                header_params,
                body,
                post_params,
                files,
                response_type,
                auth_settings,
                async_req,
                _return_http_data_only,
                collection_formats,
                _preload_content,
                _request_timeout,
                _host,
            )
        except ApiExceptionExternal as e:
            format_api_exception(
                e, method, resource_path, self.raise_structured_exception
            )


@wrapt.decorator
def make_async(_func, instance, args, kwargs):  # noqa: ARG001
    loop = asyncio.get_event_loop()
    func = functools.partial(_func, *args, **kwargs)
    return loop.run_in_executor(executor=None, func=func)


class AsyncApiClientWrapperExternal(ApiClientWrapperExternal):
    @make_async
    def call_api(self, *args, **kwargs):
        return super().call_api(*args, **kwargs)


_api_client = _ApiClient()



# coding: utf-8

"""
    Anyscale API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from anyscale_client.configuration import Configuration


class CreateComputeTemplateConfig(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'cloud_id': 'str',
        'max_workers': 'int',
        'region': 'str',
        'allowed_azs': 'list[str]',
        'head_node_type': 'ComputeNodeType',
        'worker_node_types': 'list[WorkerNodeType]',
        'aws_advanced_configurations_json': 'object',
        'aws': 'AWSNodeOptions',
        'gcp_advanced_configurations_json': 'object',
        'gcp': 'GCPNodeOptions',
        'azure': 'object',
        'maximum_uptime_minutes': 'int',
        'idle_termination_minutes': 'int'
    }

    attribute_map = {
        'cloud_id': 'cloud_id',
        'max_workers': 'max_workers',
        'region': 'region',
        'allowed_azs': 'allowed_azs',
        'head_node_type': 'head_node_type',
        'worker_node_types': 'worker_node_types',
        'aws_advanced_configurations_json': 'aws_advanced_configurations_json',
        'aws': 'aws',
        'gcp_advanced_configurations_json': 'gcp_advanced_configurations_json',
        'gcp': 'gcp',
        'azure': 'azure',
        'maximum_uptime_minutes': 'maximum_uptime_minutes',
        'idle_termination_minutes': 'idle_termination_minutes'
    }

    def __init__(self, cloud_id=None, max_workers=None, region=None, allowed_azs=None, head_node_type=None, worker_node_types=None, aws_advanced_configurations_json=None, aws=None, gcp_advanced_configurations_json=None, gcp=None, azure=None, maximum_uptime_minutes=None, idle_termination_minutes=120, local_vars_configuration=None):  # noqa: E501
        """CreateComputeTemplateConfig - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._cloud_id = None
        self._max_workers = None
        self._region = None
        self._allowed_azs = None
        self._head_node_type = None
        self._worker_node_types = None
        self._aws_advanced_configurations_json = None
        self._aws = None
        self._gcp_advanced_configurations_json = None
        self._gcp = None
        self._azure = None
        self._maximum_uptime_minutes = None
        self._idle_termination_minutes = None
        self.discriminator = None

        self.cloud_id = cloud_id
        if max_workers is not None:
            self.max_workers = max_workers
        self.region = region
        if allowed_azs is not None:
            self.allowed_azs = allowed_azs
        self.head_node_type = head_node_type
        self.worker_node_types = worker_node_types
        if aws_advanced_configurations_json is not None:
            self.aws_advanced_configurations_json = aws_advanced_configurations_json
        if aws is not None:
            self.aws = aws
        if gcp_advanced_configurations_json is not None:
            self.gcp_advanced_configurations_json = gcp_advanced_configurations_json
        if gcp is not None:
            self.gcp = gcp
        if azure is not None:
            self.azure = azure
        if maximum_uptime_minutes is not None:
            self.maximum_uptime_minutes = maximum_uptime_minutes
        if idle_termination_minutes is not None:
            self.idle_termination_minutes = idle_termination_minutes

    @property
    def cloud_id(self):
        """Gets the cloud_id of this CreateComputeTemplateConfig.  # noqa: E501

        The ID of the Anyscale cloud to use for launching sessions.  # noqa: E501

        :return: The cloud_id of this CreateComputeTemplateConfig.  # noqa: E501
        :rtype: str
        """
        return self._cloud_id

    @cloud_id.setter
    def cloud_id(self, cloud_id):
        """Sets the cloud_id of this CreateComputeTemplateConfig.

        The ID of the Anyscale cloud to use for launching sessions.  # noqa: E501

        :param cloud_id: The cloud_id of this CreateComputeTemplateConfig.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and cloud_id is None:  # noqa: E501
            raise ValueError("Invalid value for `cloud_id`, must not be `None`")  # noqa: E501

        self._cloud_id = cloud_id

    @property
    def max_workers(self):
        """Gets the max_workers of this CreateComputeTemplateConfig.  # noqa: E501

        Desired limit on total running workers for this session.  # noqa: E501

        :return: The max_workers of this CreateComputeTemplateConfig.  # noqa: E501
        :rtype: int
        """
        return self._max_workers

    @max_workers.setter
    def max_workers(self, max_workers):
        """Sets the max_workers of this CreateComputeTemplateConfig.

        Desired limit on total running workers for this session.  # noqa: E501

        :param max_workers: The max_workers of this CreateComputeTemplateConfig.  # noqa: E501
        :type: int
        """

        self._max_workers = max_workers

    @property
    def region(self):
        """Gets the region of this CreateComputeTemplateConfig.  # noqa: E501

        The region to launch sessions in, e.g. \"us-west-2\".  # noqa: E501

        :return: The region of this CreateComputeTemplateConfig.  # noqa: E501
        :rtype: str
        """
        return self._region

    @region.setter
    def region(self, region):
        """Sets the region of this CreateComputeTemplateConfig.

        The region to launch sessions in, e.g. \"us-west-2\".  # noqa: E501

        :param region: The region of this CreateComputeTemplateConfig.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and region is None:  # noqa: E501
            raise ValueError("Invalid value for `region`, must not be `None`")  # noqa: E501

        self._region = region

    @property
    def allowed_azs(self):
        """Gets the allowed_azs of this CreateComputeTemplateConfig.  # noqa: E501

        The availability zones that sessions are allowed to be launched in, e.g. \"us-west-2a\". If not specified, any AZ may be used.  # noqa: E501

        :return: The allowed_azs of this CreateComputeTemplateConfig.  # noqa: E501
        :rtype: list[str]
        """
        return self._allowed_azs

    @allowed_azs.setter
    def allowed_azs(self, allowed_azs):
        """Sets the allowed_azs of this CreateComputeTemplateConfig.

        The availability zones that sessions are allowed to be launched in, e.g. \"us-west-2a\". If not specified, any AZ may be used.  # noqa: E501

        :param allowed_azs: The allowed_azs of this CreateComputeTemplateConfig.  # noqa: E501
        :type: list[str]
        """

        self._allowed_azs = allowed_azs

    @property
    def head_node_type(self):
        """Gets the head_node_type of this CreateComputeTemplateConfig.  # noqa: E501

        Node configuration to use for the head node.   # noqa: E501

        :return: The head_node_type of this CreateComputeTemplateConfig.  # noqa: E501
        :rtype: ComputeNodeType
        """
        return self._head_node_type

    @head_node_type.setter
    def head_node_type(self, head_node_type):
        """Sets the head_node_type of this CreateComputeTemplateConfig.

        Node configuration to use for the head node.   # noqa: E501

        :param head_node_type: The head_node_type of this CreateComputeTemplateConfig.  # noqa: E501
        :type: ComputeNodeType
        """
        if self.local_vars_configuration.client_side_validation and head_node_type is None:  # noqa: E501
            raise ValueError("Invalid value for `head_node_type`, must not be `None`")  # noqa: E501

        self._head_node_type = head_node_type

    @property
    def worker_node_types(self):
        """Gets the worker_node_types of this CreateComputeTemplateConfig.  # noqa: E501

        A list of node types to use for worker nodes.   # noqa: E501

        :return: The worker_node_types of this CreateComputeTemplateConfig.  # noqa: E501
        :rtype: list[WorkerNodeType]
        """
        return self._worker_node_types

    @worker_node_types.setter
    def worker_node_types(self, worker_node_types):
        """Sets the worker_node_types of this CreateComputeTemplateConfig.

        A list of node types to use for worker nodes.   # noqa: E501

        :param worker_node_types: The worker_node_types of this CreateComputeTemplateConfig.  # noqa: E501
        :type: list[WorkerNodeType]
        """
        if self.local_vars_configuration.client_side_validation and worker_node_types is None:  # noqa: E501
            raise ValueError("Invalid value for `worker_node_types`, must not be `None`")  # noqa: E501

        self._worker_node_types = worker_node_types

    @property
    def aws_advanced_configurations_json(self):
        """Gets the aws_advanced_configurations_json of this CreateComputeTemplateConfig.  # noqa: E501

        The advanced configuration json that we pass directly AWS APIs when launching an instance. We may do some validation on this json and reject the json if it is using a configuration that Anyscale does not support.  # noqa: E501

        :return: The aws_advanced_configurations_json of this CreateComputeTemplateConfig.  # noqa: E501
        :rtype: object
        """
        return self._aws_advanced_configurations_json

    @aws_advanced_configurations_json.setter
    def aws_advanced_configurations_json(self, aws_advanced_configurations_json):
        """Sets the aws_advanced_configurations_json of this CreateComputeTemplateConfig.

        The advanced configuration json that we pass directly AWS APIs when launching an instance. We may do some validation on this json and reject the json if it is using a configuration that Anyscale does not support.  # noqa: E501

        :param aws_advanced_configurations_json: The aws_advanced_configurations_json of this CreateComputeTemplateConfig.  # noqa: E501
        :type: object
        """

        self._aws_advanced_configurations_json = aws_advanced_configurations_json

    @property
    def aws(self):
        """Gets the aws of this CreateComputeTemplateConfig.  # noqa: E501

        DEPRECATED: Please provide instance configuration in the `aws_advanced_configurations_json` field. Fields specific to AWS node types.  # noqa: E501

        :return: The aws of this CreateComputeTemplateConfig.  # noqa: E501
        :rtype: AWSNodeOptions
        """
        return self._aws

    @aws.setter
    def aws(self, aws):
        """Sets the aws of this CreateComputeTemplateConfig.

        DEPRECATED: Please provide instance configuration in the `aws_advanced_configurations_json` field. Fields specific to AWS node types.  # noqa: E501

        :param aws: The aws of this CreateComputeTemplateConfig.  # noqa: E501
        :type: AWSNodeOptions
        """

        self._aws = aws

    @property
    def gcp_advanced_configurations_json(self):
        """Gets the gcp_advanced_configurations_json of this CreateComputeTemplateConfig.  # noqa: E501

        The advanced configuration json that we pass directly GCP APIs when launching an instance. We may do some validation on this json and reject the json if it is using a configuration that Anyscale does not support.  # noqa: E501

        :return: The gcp_advanced_configurations_json of this CreateComputeTemplateConfig.  # noqa: E501
        :rtype: object
        """
        return self._gcp_advanced_configurations_json

    @gcp_advanced_configurations_json.setter
    def gcp_advanced_configurations_json(self, gcp_advanced_configurations_json):
        """Sets the gcp_advanced_configurations_json of this CreateComputeTemplateConfig.

        The advanced configuration json that we pass directly GCP APIs when launching an instance. We may do some validation on this json and reject the json if it is using a configuration that Anyscale does not support.  # noqa: E501

        :param gcp_advanced_configurations_json: The gcp_advanced_configurations_json of this CreateComputeTemplateConfig.  # noqa: E501
        :type: object
        """

        self._gcp_advanced_configurations_json = gcp_advanced_configurations_json

    @property
    def gcp(self):
        """Gets the gcp of this CreateComputeTemplateConfig.  # noqa: E501

        DEPRECATED: Please provide instance configuration in the `gcp_advanced_configurations_json` field. Fields specific to GCP node types.  # noqa: E501

        :return: The gcp of this CreateComputeTemplateConfig.  # noqa: E501
        :rtype: GCPNodeOptions
        """
        return self._gcp

    @gcp.setter
    def gcp(self, gcp):
        """Sets the gcp of this CreateComputeTemplateConfig.

        DEPRECATED: Please provide instance configuration in the `gcp_advanced_configurations_json` field. Fields specific to GCP node types.  # noqa: E501

        :param gcp: The gcp of this CreateComputeTemplateConfig.  # noqa: E501
        :type: GCPNodeOptions
        """

        self._gcp = gcp

    @property
    def azure(self):
        """Gets the azure of this CreateComputeTemplateConfig.  # noqa: E501

        DEPRECATED: We don't currently support azure. Fields specific to Azure node types.  # noqa: E501

        :return: The azure of this CreateComputeTemplateConfig.  # noqa: E501
        :rtype: object
        """
        return self._azure

    @azure.setter
    def azure(self, azure):
        """Sets the azure of this CreateComputeTemplateConfig.

        DEPRECATED: We don't currently support azure. Fields specific to Azure node types.  # noqa: E501

        :param azure: The azure of this CreateComputeTemplateConfig.  # noqa: E501
        :type: object
        """

        self._azure = azure

    @property
    def maximum_uptime_minutes(self):
        """Gets the maximum_uptime_minutes of this CreateComputeTemplateConfig.  # noqa: E501

        If set to a positive number, Anyscale will terminate the cluster this many minutes after cluster start.  # noqa: E501

        :return: The maximum_uptime_minutes of this CreateComputeTemplateConfig.  # noqa: E501
        :rtype: int
        """
        return self._maximum_uptime_minutes

    @maximum_uptime_minutes.setter
    def maximum_uptime_minutes(self, maximum_uptime_minutes):
        """Sets the maximum_uptime_minutes of this CreateComputeTemplateConfig.

        If set to a positive number, Anyscale will terminate the cluster this many minutes after cluster start.  # noqa: E501

        :param maximum_uptime_minutes: The maximum_uptime_minutes of this CreateComputeTemplateConfig.  # noqa: E501
        :type: int
        """

        self._maximum_uptime_minutes = maximum_uptime_minutes

    @property
    def idle_termination_minutes(self):
        """Gets the idle_termination_minutes of this CreateComputeTemplateConfig.  # noqa: E501

        If set to a positive number, Anyscale will terminate the cluster this many minutes after the cluster is idle. Idle time is defined as the time during which a Cluster is not running a user command or a Ray driver. Time spent running commands on Jupyter or ssh is still considered 'idle'. To disable, set this field to 0.  # noqa: E501

        :return: The idle_termination_minutes of this CreateComputeTemplateConfig.  # noqa: E501
        :rtype: int
        """
        return self._idle_termination_minutes

    @idle_termination_minutes.setter
    def idle_termination_minutes(self, idle_termination_minutes):
        """Sets the idle_termination_minutes of this CreateComputeTemplateConfig.

        If set to a positive number, Anyscale will terminate the cluster this many minutes after the cluster is idle. Idle time is defined as the time during which a Cluster is not running a user command or a Ray driver. Time spent running commands on Jupyter or ssh is still considered 'idle'. To disable, set this field to 0.  # noqa: E501

        :param idle_termination_minutes: The idle_termination_minutes of this CreateComputeTemplateConfig.  # noqa: E501
        :type: int
        """
        if (self.local_vars_configuration.client_side_validation and
                idle_termination_minutes is not None and idle_termination_minutes < 0):  # noqa: E501
            raise ValueError("Invalid value for `idle_termination_minutes`, must be a value greater than or equal to `0`")  # noqa: E501

        self._idle_termination_minutes = idle_termination_minutes

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, CreateComputeTemplateConfig):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, CreateComputeTemplateConfig):
            return True

        return self.to_dict() != other.to_dict()



import datetime
import logging
from typing import Optional

import click

from anyscale.project import get_project_id, get_project_session, load_project_or_throw
from anyscale.shared_anyscale_utils.util import execution_log_name
from anyscale.snapshot import copy_file
from anyscale.util import deserialize_datetime, send_json_request


logger = logging.getLogger(__file__)
logging.getLogger("botocore").setLevel(logging.CRITICAL)


@click.group("session", help="Commands for working with sessions.", hidden=True)
def session_cli() -> None:
    pass


@session_cli.command(name="logs", help="Show logs for the current session.")
@click.option("--name", help="Name of the session to run this command on", default=None)
@click.option("--command-id", help="ID of the command to get logs for", default=None)
def session_logs(name: Optional[str], command_id: Optional[int]) -> None:
    project_definition = load_project_or_throw()
    project_id = get_project_id(project_definition.root)
    # If the command_id is not specified, determine it by getting the
    # last run command from the active session.
    if not command_id:
        session = get_project_session(project_id, name)
        resp = send_json_request(
            "/api/v2/session_commands/?session_id={}".format(session["id"]), {}
        )
        # Search for latest run command
        last_created_at = datetime.datetime.min
        last_created_at = last_created_at.replace(tzinfo=datetime.timezone.utc)
        for command in resp["results"]:
            created_at = deserialize_datetime(command["created_at"])
            if created_at > last_created_at:
                last_created_at = created_at
                command_id = command["id"]
        if not command_id:
            raise click.ClickException(
                "No comand was run yet on the latest active session {}".format(
                    session["name"]
                )
            )
    resp_out = send_json_request(
        "/api/v2/session_commands/{session_command_id}/execution_logs".format(
            session_command_id=command_id
        ),
        {"log_type": "out", "start_line": 0, "end_line": 1000000000},
    )
    resp_err = send_json_request(
        "/api/v2/session_commands/{session_command_id}/execution_logs".format(
            session_command_id=command_id
        ),
        {"log_type": "err", "start_line": 0, "end_line": 1000000000},
    )
    # TODO(pcm): We should have more options here in the future
    # (e.g. show only stdout or stderr, show only the tail, etc).
    print("stdout:")
    print(resp_out["result"]["lines"])
    print("stderr:")
    print(resp_err["result"]["lines"])


# Note: These functions are not normally updated, please cc yiran or ijrsvt on changes.
@session_cli.command(
    name="upload_command_logs", help="Upload logs for a command.", hidden=True
)
@click.option(
    "--command-id", help="ID of the command to upload logs for", type=str, default=None
)
def session_upload_command_logs(command_id: Optional[str],) -> None:
    session_upload_command_logs_impl(command_id)


# session_upload_command_logs_impl is used by:
# - upload_command_logs CLI command
# - web terminal command persister queue
# a separate function is needed to be able to call
# the logic from code rather just CLI.
def session_upload_command_logs_impl(
    command_id: Optional[str],
    cli_token: Optional[str] = None,
    host: Optional[str] = None,
) -> None:
    resp = send_json_request(
        "/api/v2/session_commands/{session_command_id}/upload_logs".format(
            session_command_id=command_id
        ),
        {},
        method="POST",
        cli_token=cli_token,
        host=host,
    )
    assert resp["result"]["session_command_id"] == command_id

    allowed_sources = [
        execution_log_name(command_id) + ".out",
        execution_log_name(command_id) + ".err",
    ]

    for source, target in resp["result"]["locations"].items():
        if source in allowed_sources:
            copy_file(True, source, target, download=False)


# Note: These functions are not normally updated, please cc yiran or ijrsvt on changes.
@session_cli.command(
    name="finish_command", help="Finish executing a command.", hidden=True
)
@click.option(
    "--command-id", help="ID of the command to finish", type=str, required=True
)
@click.option(
    "--stop", help="Stop session after command finishes executing.", is_flag=True
)
@click.option(
    "--terminate",
    help="Terminate session after command finishes executing.",
    is_flag=True,
)
def session_finish_command(command_id: str, stop: bool, terminate: bool) -> None:
    with open(execution_log_name(command_id) + ".status") as f:
        status_code = int(f.read().strip())
    send_json_request(
        f"/api/v2/session_commands/{command_id}/finish",
        {"status_code": status_code, "stop": stop, "terminate": terminate},
        method="POST",
    )


@session_cli.command(
    name="web_terminal_server", help="Start the web terminal server", hidden=True
)
@click.option(
    "--deploy-environment",
    help="Anyscale deployment type (development, test, staging, production)",
    type=str,
    required=True,
)
@click.option(
    "--use-debugger", help="Activate the Anyscale debugger.", is_flag=True,
)
@click.option(
    "--cli-token",
    help="Anyscale cli token used to instantiate anyscale openapi",
    type=str,
    required=True,
)
@click.option(
    "--host",
    help="Anyscale host used to instantiate anyscale openapi (console.anyscale.com for example)",
    type=str,
    required=True,
)
@click.option(
    "--working-dir",
    help="The working directory for this anyscale session. The webterminal will be opened from this directory.",
    type=str,
    required=True,
)
@click.option(
    "--session-id", help="The session id of this web terminal", type=str, required=True,
)
def web_terminal_server(  # noqa: PLR0913
    deploy_environment: str,
    use_debugger: bool,
    cli_token: str,
    host: str,
    working_dir: str,
    session_id: str,
) -> None:
    from anyscale.webterminal.webterminal import main

    main(deploy_environment, use_debugger, cli_token, host, working_dir, session_id)



import json
import random
import re
import string
import sys
from typing import Any, Callable, Dict, List, Optional
from unittest.mock import Mock, patch

import boto3
from moto import mock_cloudformation, mock_ec2, mock_efs, mock_iam, mock_s3
import pytest

from anyscale.cloud_resource import (
    associate_subnets_with_azs,
    aws_subnet_has_enough_capacity,
    verify_aws_cloudformation_stack,
    verify_aws_efs,
    verify_aws_iam_roles,
    verify_aws_s3,
    verify_aws_security_groups,
    verify_aws_subnets,
    verify_aws_vpc,
)
from anyscale.conf import ANYSCALE_IAM_ROLE_NAME
from anyscale.shared_anyscale_utils.aws import AwsRoleArn
from anyscale.shared_anyscale_utils.conf import ANYSCALE_HOST
from frontend.cli.anyscale.aws_iam_policies import (
    AMAZON_ECR_READONLY_ACCESS_POLICY_ARN,
    AMAZON_S3_FULL_ACCESS_POLICY_ARN,
    ANYSCALE_IAM_PERMISSIONS_EC2_INITIAL_RUN,
    ANYSCALE_IAM_PERMISSIONS_EC2_STEADY_STATE,
    get_anyscale_aws_iam_assume_role_policy,
)
from frontend.cli.anyscale.cli_logger import BlockLogger
from frontend.cli.anyscale.client.openapi_client.models.create_cloud_resource import (
    CreateCloudResource,
)


DEFAULT_RAY_IAM_ROLE = "ray-autoscaler-v1"


def _create_roles(anyscale_account: str):
    iam = boto3.client("iam")
    suffix = "".join(random.choices(string.ascii_lowercase, k=10))
    control_plane_role = iam.create_role(
        RoleName=ANYSCALE_IAM_ROLE_NAME + suffix,
        AssumeRolePolicyDocument=json.dumps(
            get_anyscale_aws_iam_assume_role_policy(
                anyscale_aws_account=anyscale_account
            )
        ),
    )["Role"]["Arn"]

    # Create & Configure the Dataplane Role
    dataplane_role_name = DEFAULT_RAY_IAM_ROLE + suffix
    dataplane_role = iam.create_role(
        RoleName=dataplane_role_name, AssumeRolePolicyDocument="{}"
    )["Role"]["Arn"]

    iam.attach_role_policy(
        RoleName=dataplane_role_name, PolicyArn=AMAZON_ECR_READONLY_ACCESS_POLICY_ARN
    )
    iam.attach_role_policy(
        RoleName=dataplane_role_name, PolicyArn=AMAZON_S3_FULL_ACCESS_POLICY_ARN
    )
    return [control_plane_role, dataplane_role]


def _delete_role(role_arn: str) -> None:
    iam = boto3.client("iam")
    role_name = AwsRoleArn.from_string(role_arn).to_role_name()

    policies = iam.list_attached_role_policies(RoleName=role_name)["AttachedPolicies"]

    for policy in policies:
        iam.detach_role_policy(RoleName=role_name, PolicyArn=policy["PolicyArn"])

    iam.delete_role(RoleName=role_name)


def generate_cloud_resource_mock_aws() -> CreateCloudResource:
    return CreateCloudResource(
        aws_vpc_id="fake_aws_vpc_id",
        aws_subnet_ids=["fake_aws_subnet_id_0", "fake_aws_subnet_id_1"],
        aws_iam_role_arns=[
            "arn:aws:iam::123:role/mock_anyscale_role",
            "arn:aws:iam::123:role/mock_dataplane_role",
        ],
        aws_security_groups=["fake_aws_security_group_0"],
        aws_s3_id="fake_aws_s3_id",
        aws_efs_id="fake_aws_efs_id",
        aws_cloudformation_stack_id="fake_aws_cloudformation_stack_id",
    )


def _attach_iam_roles_to_role(role_arn: str) -> None:
    iam = boto3.client("iam")
    role_name = AwsRoleArn.from_string(role_arn).to_role_name()

    steady_state = iam.create_policy(
        PolicyName="".join(random.choices(string.ascii_lowercase, k=10)),
        PolicyDocument=json.dumps(ANYSCALE_IAM_PERMISSIONS_EC2_STEADY_STATE),
    )["Policy"]["Arn"]
    iam.attach_role_policy(RoleName=role_name, PolicyArn=steady_state)

    initial_run = iam.create_policy(
        PolicyName="".join(random.choices(string.ascii_lowercase, k=10)),
        PolicyDocument=json.dumps(ANYSCALE_IAM_PERMISSIONS_EC2_INITIAL_RUN),
    )["Policy"]["Arn"]
    iam.attach_role_policy(RoleName=role_name, PolicyArn=initial_run)


def _attach_inline_iam_policy_to_role(role_arn: str) -> None:
    iam = boto3.client("iam")
    role_name = AwsRoleArn.from_string(role_arn).to_role_name()

    iam.put_role_policy(
        RoleName=role_name,
        PolicyName="".join(random.choices(string.ascii_lowercase, k=10)),
        PolicyDocument=json.dumps(ANYSCALE_IAM_PERMISSIONS_EC2_STEADY_STATE),
    )
    iam.put_role_policy(
        RoleName=role_name,
        PolicyName="".join(random.choices(string.ascii_lowercase, k=10)),
        PolicyDocument=json.dumps(ANYSCALE_IAM_PERMISSIONS_EC2_INITIAL_RUN),
    )


def _create_instance_profile_for_role(
    role_arn: str, instance_profile_name: Optional[str]
) -> None:
    role_name = AwsRoleArn.from_string(role_arn).to_role_name()
    if instance_profile_name is None:
        instance_profile_name = role_name

    iam = boto3.client("iam")
    iam.create_instance_profile(InstanceProfileName=instance_profile_name)
    iam.add_role_to_instance_profile(
        RoleName=role_name, InstanceProfileName=instance_profile_name,
    )


def _create_vpc(cidr_block: str = "10.0.0.0/16", region: str = "us-west-2") -> str:
    ec2 = boto3.client("ec2", region_name=region)
    vpc = ec2.create_vpc(CidrBlock=cidr_block)
    return vpc["Vpc"]["VpcId"]


def _create_subnets(
    vpc_id: str, region: str, num_subnets: int = 2, assign_public_ip: bool = False
) -> List[str]:
    ec2 = boto3.client("ec2", region_name=region)
    subnet_ids = []
    for i in range(num_subnets):
        subnet = ec2.create_subnet(
            VpcId=vpc_id,
            CidrBlock=f"10.0.{i}.0/24",
            AvailabilityZone=f"{region}{string.ascii_lowercase[i]}",
        )
        subnet_id = subnet["Subnet"]["SubnetId"]
        subnet_ids.append(subnet_id)
        # MapPublicIpOnLaunch is by default False.
        if assign_public_ip:
            ec2.modify_subnet_attribute(
                SubnetId=subnet_id, MapPublicIpOnLaunch={"Value": True},
            )

    return subnet_ids


def _create_security_group(region: str) -> str:
    ec2 = boto3.client("ec2", region_name=region)
    vpc_id = _create_vpc(region=region)

    security_group = ec2.create_security_group(
        VpcId=vpc_id, GroupName="mock_security_group", Description="mock_security_group"
    )
    security_group_id = security_group["GroupId"]
    ec2.authorize_security_group_ingress(
        GroupId=security_group_id,
        IpPermissions=[
            {
                "IpProtocol": "-1",
                "FromPort": -1,
                "ToPort": -1,
                "UserIdGroupPairs": [{"GroupId": security_group_id,}],
            }
        ],
    )

    return security_group_id


def _create_security_group_inbound_permission_expected(region: str) -> str:
    ec2 = boto3.client("ec2", region_name=region)
    vpc_id = _create_vpc(region=region)

    security_group = ec2.create_security_group(
        VpcId=vpc_id, GroupName="mock_security_group", Description="mock_security_group"
    )
    security_group_id = security_group["GroupId"]
    ec2.authorize_security_group_ingress(
        GroupId=security_group_id,
        IpPermissions=[
            {
                "IpProtocol": "tcp",  # use tcp instead of -1 because of moto impl issue
                "FromPort": 443,
                "ToPort": 443,
                "UserIdGroupPairs": [{"GroupId": security_group_id}],
            },
            {
                "IpProtocol": "tcp",  # use tcp instead of -1 because of moto impl issue
                "FromPort": 22,
                "ToPort": 22,
                "UserIdGroupPairs": [{"GroupId": security_group_id}],
            },
            {
                "IpProtocol": "-1",  # -1 is needed for check here: https://github.com/anyscale/product/blob/master/frontend/cli/anyscale/cloud_resource.py#L378
                "FromPort": -1,
                "ToPort": -1,
                "UserIdGroupPairs": [{"GroupId": security_group_id}],
            },
        ],
    )
    return security_group_id


def _create_security_group_inbound_permission_excessive_permission(region: str) -> str:
    ec2 = boto3.client("ec2", region_name=region)
    vpc_id = _create_vpc(region=region)

    security_group = ec2.create_security_group(
        VpcId=vpc_id, GroupName="mock_security_group", Description="mock_security_group"
    )
    security_group_id = security_group["GroupId"]
    ec2.authorize_security_group_ingress(
        GroupId=security_group_id,
        IpPermissions=[
            {
                "IpProtocol": "tcp",  # use tcp instead of -1 because of moto impl issue
                "FromPort": 443,
                "ToPort": 443,
                "UserIdGroupPairs": [{"GroupId": security_group_id}],
            },
            {
                "IpProtocol": "tcp",  # use tcp instead of -1 because of moto impl issue
                "FromPort": 22,
                "ToPort": 22,
                "UserIdGroupPairs": [{"GroupId": security_group_id}],
            },
            {
                "IpProtocol": "tcp",  # use tcp instead of -1 because of moto impl issue
                "FromPort": 123,
                "ToPort": 123,
                "UserIdGroupPairs": [{"GroupId": security_group_id}],
            },
            {
                "IpProtocol": "-1",  # -1 is needed for check here: https://github.com/anyscale/product/blob/master/frontend/cli/anyscale/cloud_resource.py#L378
                "FromPort": -1,
                "ToPort": -1,
                "UserIdGroupPairs": [{"GroupId": security_group_id}],
            },
        ],
    )
    return security_group_id


def _create_bucket(bucket_name: str = "mock_bucket", cors_rules: Any = None) -> Any:
    s3 = boto3.client("s3")
    s3.create_bucket(Bucket=bucket_name)
    if cors_rules:
        s3.put_bucket_cors(
            Bucket=bucket_name, CORSConfiguration={"CORSRules": cors_rules}
        )
    return boto3.resource("s3").Bucket(bucket_name)


def _create_efs(region: str, subnet_ids: List[str], security_group_ids: List[str]):
    efs = boto3.client("efs", region_name=region)
    create_resp = efs.create_file_system(CreationToken="mock_efs")
    efs_id = create_resp["FileSystemId"]

    for subnet_id in subnet_ids:
        efs.create_mount_target(
            FileSystemId=efs_id, SubnetId=subnet_id, SecurityGroups=security_group_ids,
        )
    return efs_id


def _create_cloudformation_stack(region: str) -> str:
    cloudformation = boto3.client("cloudformation", region_name=region)
    cfn = cloudformation.create_stack(
        StackName="mock_cloudformation_stack",
        TemplateBody=json.dumps(
            {
                "Resources": {
                    "mock_resource": {
                        "Type": "AWS::EC2::Instance",
                        "Properties": {
                            "InstanceType": "t2.micro",
                            "ImageId": "ami-0b69ea66ff7391e80",
                        },
                    }
                }
            }
        ),
    )

    return cfn["StackId"]


@pytest.mark.parametrize(
    ("vpc_exists", "vpc_cidr_block", "expected_result", "expected_log_message"),
    [
        pytest.param(False, "0.0.0.0/0", False, r"Could not find"),
        # Happy sizes
        pytest.param(True, "0.0.0.0/16", True, r"verification succeeded."),
        pytest.param(True, "0.0.0.0/20", True, r"verification succeeded."),
        # Warn sizes
        pytest.param(
            True, "0.0.3.0/21", True, r"but this vpc only supports up to \d+ addresses",
        ),
        pytest.param(
            True, "0.0.4.0/24", True, r"but this vpc only supports up to \d+ addresses",
        ),
        # Error sizes
        pytest.param(
            True,
            "0.0.4.0/25",
            False,
            r"Please reach out to support if this is an issue!",
        ),
    ],
)
@mock_ec2
def test_verify_aws_vpc(
    capsys,
    vpc_exists: bool,
    vpc_cidr_block: str,
    expected_result: bool,
    expected_log_message: str,
):
    cloud_resource_mock = generate_cloud_resource_mock_aws()
    region = "us-west-2"
    if vpc_exists:
        vpc_id = _create_vpc(cidr_block=vpc_cidr_block, region=region)
        cloud_resource_mock.aws_vpc_id = vpc_id

    result = verify_aws_vpc(
        cloud_resource=cloud_resource_mock,
        boto3_session=boto3.Session(region_name=region),
        logger=BlockLogger(),
    )
    assert result == expected_result

    stdout, stderr = capsys.readouterr()
    sys.stdout.write(stdout)
    sys.stderr.write(stderr)

    if expected_log_message:
        assert re.search(expected_log_message, stderr)


@pytest.mark.parametrize(
    ("is_private_network"), [False, True],
)
@mock_ec2
def test_verify_aws_subnets(capsys, is_private_network: bool):
    cloud_resource_mock = generate_cloud_resource_mock_aws()
    region = "us-west-2"
    vpc_id = _create_vpc(region=region)
    subnet_ids = _create_subnets(
        vpc_id=vpc_id, region=region, assign_public_ip=not is_private_network
    )

    cloud_resource_mock.aws_vpc_id = vpc_id
    cloud_resource_mock.aws_subnet_ids = subnet_ids

    result = verify_aws_subnets(
        cloud_resource=cloud_resource_mock,
        region=region,
        is_private_network=is_private_network,
        logger=BlockLogger(),
    )

    _, stderr = capsys.readouterr()

    assert result
    assert not re.search(
        r"does not have the 'Auto-assign Public IP' option enabled", stderr
    )
    assert re.search(r"verification succeeded.", stderr)


@mock_ec2
def test_verify_aws_subnets_warn_about_not_assigning_public_ip_in_public_subnets(
    capsys,
):
    cloud_resource_mock = generate_cloud_resource_mock_aws()
    region = "us-west-2"
    vpc_id = _create_vpc(region=region)
    subnet_ids = _create_subnets(vpc_id=vpc_id, region=region, assign_public_ip=False)

    cloud_resource_mock.aws_vpc_id = vpc_id
    cloud_resource_mock.aws_subnet_ids = subnet_ids

    result = verify_aws_subnets(
        cloud_resource=cloud_resource_mock,
        region=region,
        is_private_network=False,
        logger=BlockLogger(),
    )

    _, stderr = capsys.readouterr()

    assert result
    assert re.search(
        r"does not have the 'Auto-assign Public IP' option enabled", stderr
    )
    assert re.search(r"verification succeeded.", stderr)


@mock_ec2
def test_verify_aws_subnets_vpc_subnet_mismatch(capsys):
    cloud_resource_mock = generate_cloud_resource_mock_aws()

    region = "us-west-2"
    vpc_id = _create_vpc(region=region)
    subnet_ids = _create_subnets(vpc_id=vpc_id, region=region, num_subnets=2)

    irrevalent_vpc_id = _create_vpc(region=region)

    cloud_resource_mock.aws_vpc_id = irrevalent_vpc_id
    cloud_resource_mock.aws_subnet_ids = subnet_ids

    result = verify_aws_subnets(
        cloud_resource=cloud_resource_mock,
        region=region,
        is_private_network=False,
        logger=BlockLogger(),
    )

    _, stderr = capsys.readouterr()

    assert not result
    assert re.search(r"is not in a vpc of this cloud.", stderr)


@mock_ec2
def test_verify_aws_subnets_insufficient_subnets(capsys):
    cloud_resource_mock = generate_cloud_resource_mock_aws()

    region = "us-west-2"
    vpc_id = _create_vpc(region=region)
    subnet_ids = _create_subnets(vpc_id=vpc_id, region=region, num_subnets=1)

    cloud_resource_mock.aws_vpc_id = vpc_id
    cloud_resource_mock.aws_subnet_ids = subnet_ids
    cloud_resource_mock.aws_subnet_ids_with_availability_zones = None

    result = verify_aws_subnets(
        cloud_resource=cloud_resource_mock,
        region=region,
        is_private_network=False,
        logger=BlockLogger(),
    )
    assert result is False
    _, stderr = capsys.readouterr()
    assert re.search(r"Need at least 2 subnets for a cloud.", stderr)


@mock_ec2
def test_verify_aws_subnets_no_subnets():
    cloud_resource_mock = generate_cloud_resource_mock_aws()

    cloud_resource_mock.aws_subnet_ids = None
    cloud_resource_mock.aws_subnet_ids_with_availability_zones = None

    result = verify_aws_subnets(
        cloud_resource=cloud_resource_mock,
        region="us-west-2",
        is_private_network=False,
        logger=BlockLogger(),
    )
    assert result is False


@pytest.mark.parametrize(
    ("subnet_cidr", "has_capacity", "expected_log_message"),
    [
        # Happy sizes
        ("10.0.0.0/16", True, None),
        ("0.0.0.0/0", True, None),
        ("0.0.1.0/24", True, None),
        # Warn sizes
        ("0.0.2.0/25", True, r"but this subnet only supports up to \d+ addresses"),
        ("0.0.3.0/28", True, r"but this subnet only supports up to \d+ addresses"),
        # Error sizes
        ("0.0.3.0/29", False, r"Please reach out to support if this is an issue!"),
    ],
)
def test_aws_subnet_has_enough_capacity(
    subnet_cidr, has_capacity, expected_log_message, capsys
):
    subnet = Mock()
    subnet.id = "vpc-fake_id"
    subnet.cidr_block = subnet_cidr

    assert aws_subnet_has_enough_capacity(subnet, BlockLogger()) == has_capacity

    stdout, stderr = capsys.readouterr()
    sys.stdout.write(stdout)
    sys.stderr.write(stderr)

    if expected_log_message:
        assert re.search(expected_log_message, stderr)


@mock_iam
@pytest.mark.parametrize(
    ("iam_roles_mock_fn", "expect_warning"),
    [
        pytest.param(lambda _: None, True, id="No Policies"),
        pytest.param(_attach_iam_roles_to_role, False, id="Separate Policies"),
        pytest.param(_attach_inline_iam_policy_to_role, False, id="Inline Policies"),
    ],
)
def test_verify_aws_iam_roles(
    iam_roles_mock_fn: Callable[[str], None], expect_warning: bool
):
    cloud_resource_mock = generate_cloud_resource_mock_aws()
    cp_role_arn, dp_role_arn = _create_roles("fake_anyscale_aws_account")
    _create_instance_profile_for_role(dp_role_arn, None)
    cloud_resource_mock.aws_iam_role_arns = [cp_role_arn, dp_role_arn]
    iam_roles_mock_fn(cp_role_arn)
    mock_logger = Mock()
    assert verify_aws_iam_roles(
        cloud_resource=cloud_resource_mock,
        boto3_session=boto3.Session(),
        anyscale_aws_account="fake_anyscale_aws_account",
        logger=mock_logger,
    )

    if expect_warning:
        assert len(mock_logger.warning.mock_calls) == 1
    else:
        mock_logger.warning.assert_not_called()

    assert len(mock_logger.info.mock_calls) == 2


@mock_iam
@pytest.mark.parametrize(
    ("cp_role_exist", "dp_role_exist"),
    [pytest.param(True, False), pytest.param(False, True),],
)
def test_verify_aws_iam_roles_not_exist(cp_role_exist: bool, dp_role_exist: bool):
    cloud_resource_mock = generate_cloud_resource_mock_aws()
    cp_role_arn, dp_role_arn = _create_roles("fake_anyscale_aws_account")

    if not cp_role_exist:
        _delete_role(cp_role_arn)

    if not dp_role_exist:
        _delete_role(dp_role_arn)

    assert not verify_aws_iam_roles(
        cloud_resource=cloud_resource_mock,
        boto3_session=boto3.Session(),
        anyscale_aws_account="fake_anyscale_aws_account",
        logger=BlockLogger(),
    )


@mock_iam
def test_verify_aws_iam_roles_instance_profiles():
    """Tests that the dataplane role only passes validation if it has
    an InstanceProfile with the same name as the Role."""
    cloud_resource_mock = generate_cloud_resource_mock_aws()
    cp_role_arn, dp_role_arn = _create_roles("fake_anyscale_aws_account")
    cloud_resource_mock.aws_iam_role_arns = [cp_role_arn, dp_role_arn]
    mock_logger = Mock()

    # No instance profile for role
    assert not verify_aws_iam_roles(
        cloud_resource=cloud_resource_mock,
        boto3_session=boto3.Session(),
        anyscale_aws_account="fake_anyscale_aws_account",
        logger=mock_logger,
    )

    # Create an instance profile for the role, but with the wrong name
    _create_instance_profile_for_role(dp_role_arn, "wrong_name")
    assert not verify_aws_iam_roles(
        cloud_resource=cloud_resource_mock,
        boto3_session=boto3.Session(),
        anyscale_aws_account="fake_anyscale_aws_account",
        logger=mock_logger,
    )

    # Create a correct insance profile with the role!
    _create_instance_profile_for_role(dp_role_arn, None)
    assert verify_aws_iam_roles(
        cloud_resource=cloud_resource_mock,
        boto3_session=boto3.Session(),
        anyscale_aws_account="fake_anyscale_aws_account",
        logger=mock_logger,
    )


@pytest.mark.parametrize(
    ("sg_exists", "sg_valid", "expected_result", "expected_log_message"),
    [
        pytest.param(
            False, False, False, r"Missing security group IDs", id="SG-Does-Not-Exist"
        ),
        pytest.param(True, False, False, r"Could not find", id="SG-Exists-Not-Valid"),
        pytest.param(True, True, True, None, id="SG-Exists-And-Valid"),
    ],
)
@mock_ec2
def test_verify_aws_security_groups_exist_and_valid(
    capsys,
    sg_exists: bool,
    sg_valid: bool,
    expected_result: bool,
    expected_log_message: str,
):
    cloud_resource_mock = generate_cloud_resource_mock_aws()

    region = "us-west-2"
    if sg_exists:
        if sg_valid:
            sg_id = _create_security_group(region)
        else:
            sg_id = "Invalid_SG"
        cloud_resource_mock.aws_security_groups = [sg_id]
    else:
        cloud_resource_mock.aws_security_groups = []

    result = verify_aws_security_groups(
        cloud_resource=cloud_resource_mock,
        boto3_session=boto3.Session(region_name=region),
        logger=BlockLogger(),
    )

    stdout, stderr = capsys.readouterr()
    sys.stdout.write(stdout)
    sys.stderr.write(stderr)

    assert result == expected_result
    if expected_log_message:
        assert re.search(expected_log_message, stderr)


@pytest.mark.parametrize(
    ("num_security_groups", "expected_result", "expected_log_message"),
    [
        pytest.param(
            2,
            True,
            r"Multiple security groups provided. Only taking the first security group and ignoring the rest.",
            id="Multiple-SG",
        ),
        pytest.param(1, True, None, id="One-SG"),
    ],
)
@mock_ec2
def test_verify_multiple_aws_security_groups_warning(
    capsys, num_security_groups: int, expected_result: bool, expected_log_message: str,
):
    cloud_resource_mock = generate_cloud_resource_mock_aws()
    region = "us-west-2"

    sg_id_1 = _create_security_group(region)
    cloud_resource_mock.aws_security_groups = [sg_id_1]

    for _ in range(num_security_groups):
        cloud_resource_mock.aws_security_groups.append(_create_security_group(region))

    result = verify_aws_security_groups(
        cloud_resource=cloud_resource_mock,
        boto3_session=boto3.Session(region_name=region),
        logger=BlockLogger(),
    )

    stdout, stderr = capsys.readouterr()
    sys.stdout.write(stdout)
    sys.stderr.write(stderr)

    assert result == expected_result
    if expected_log_message:
        assert re.search(expected_log_message, stderr)


@pytest.mark.parametrize(
    ("not_enough_permission", "excessive_permission", "expected_log_message"),
    [
        pytest.param(False, False, None, id="Expected_Permission"),
        pytest.param(
            True,
            False,
            r"does not contain inbound permission for ports",
            id="Not_Enough_Permission",
        ),
        pytest.param(
            False, True, r"allows access to more than", id="Excessive_Permission"
        ),
    ],
)
@mock_ec2
def test_verify_aws_security_groups_inbound_permissions(
    capsys,
    not_enough_permission: bool,
    excessive_permission: bool,
    expected_log_message: str,
):
    cloud_resource_mock = generate_cloud_resource_mock_aws()

    region = "us-west-2"

    if not_enough_permission:
        sg_id = _create_security_group(region)
    elif excessive_permission:
        sg_id = _create_security_group_inbound_permission_excessive_permission(region)
    else:  # expected permission
        sg_id = _create_security_group_inbound_permission_expected(region)

    cloud_resource_mock.aws_security_groups = [sg_id]

    result = verify_aws_security_groups(
        cloud_resource=cloud_resource_mock,
        boto3_session=boto3.Session(region_name=region),
        logger=BlockLogger(),
    )
    stdout, stderr = capsys.readouterr()
    sys.stdout.write(stdout)
    sys.stderr.write(stderr)

    assert result
    if expected_log_message:
        assert re.search(expected_log_message, stderr)


@pytest.mark.parametrize(
    ("s3_exists", "expected_result"),
    [
        pytest.param(False, False, id="S3-Does-Not-Exist"),
        pytest.param(True, True, id="S3-Exists"),
    ],
)
@pytest.mark.parametrize(
    ("cors_rules", "expected_warning_call_count"),
    [
        pytest.param(
            [
                {
                    "AllowedHeaders": ["*"],
                    "AllowedMethods": ["GET"],
                    "AllowedOrigins": [ANYSCALE_HOST],
                    "ExposeHeaders": [],
                }
            ],
            0,
            id="CorrectCORS",
        ),
        pytest.param(
            [
                {
                    "AllowedHeaders": ["*"],
                    "AllowedMethods": ["GET"],
                    "AllowedOrigins": ["openai.com"],
                    "ExposeHeaders": [],
                }
            ],
            1,
            id="ImproperCORS",
        ),
        pytest.param([], 2, id="NoCORS"),
    ],
)
@mock_s3
@mock_iam
def test_verify_aws_s3(
    s3_exists: bool,
    expected_result: bool,
    cors_rules: List[Dict[str, List[str]]],
    expected_warning_call_count: int,
):
    cloud_resource_mock = generate_cloud_resource_mock_aws()

    region = "us-west-2"
    if s3_exists:
        bucket = _create_bucket(cors_rules=cors_rules)
        cloud_resource_mock.aws_s3_id = bucket.name

    cp_role_arn, dp_role_arn = _create_roles("fake_anyscale_aws_account")
    cloud_resource_mock.aws_iam_role_arns = [cp_role_arn, dp_role_arn]

    mock_logger = Mock()

    with patch.multiple(
        "anyscale.cloud_resource", verify_s3_access=Mock(return_value=True),
    ):
        result = verify_aws_s3(
            cloud_resource=cloud_resource_mock,
            boto3_session=boto3.Session(region_name=region),
            logger=mock_logger,
        )

    if s3_exists:
        assert mock_logger.warning.call_count == expected_warning_call_count
    else:
        assert mock_logger.error.call_count == 1
        assert mock_logger.warning.call_count == 0

    assert result == expected_result


@pytest.mark.parametrize(
    (
        "efs_exists",
        "num_subnets_to_use",
        "use_security_group",
        "expected_result",
        "warning",
    ),
    [
        pytest.param(False, 2, True, False, False),
        pytest.param(True, 0, False, False, False),
        pytest.param(True, 1, True, True, True),
        pytest.param(True, 1, False, True, True),
        pytest.param(True, 2, True, True, False),
        pytest.param(True, 2, False, True, True),
    ],
)
@mock_efs
@mock_ec2
def test_verify_aws_efs(
    capsys,
    efs_exists: bool,
    num_subnets_to_use: int,
    use_security_group: bool,
    expected_result: bool,
    warning: bool,
):
    cloud_resource_mock = generate_cloud_resource_mock_aws()

    region = "us-west-2"
    if efs_exists:
        vpc_id = _create_vpc(region=region)
        subnet_ids = _create_subnets(vpc_id, region, num_subnets=2)
        assert len(subnet_ids) == 2
        security_group_id = _create_security_group(region)
        subnet_ids_in_efs = subnet_ids[:num_subnets_to_use]
        security_groups_in_efs = [security_group_id] if use_security_group else []
        efs_id = _create_efs(region, subnet_ids_in_efs, security_groups_in_efs)
        cloud_resource_mock.aws_subnet_ids = subnet_ids
        cloud_resource_mock.aws_security_groups = [security_group_id]
        cloud_resource_mock.aws_efs_id = efs_id

    result = verify_aws_efs(
        cloud_resource=cloud_resource_mock,
        boto3_session=boto3.Session(region_name=region),
        logger=BlockLogger(),
    )
    if warning:
        _, stderr = capsys.readouterr()
        assert re.search(r"does not contain network interface with subnet id", stderr)
    assert result == expected_result


@pytest.mark.parametrize(
    ("cloudformation_stack_exists", "expected_result"),
    [pytest.param(False, False), pytest.param(True, True)],
)
@mock_cloudformation
def test_verify_aws_cloudformation_stack(
    cloudformation_stack_exists: bool, expected_result: bool
):
    cloud_resource_mock = generate_cloud_resource_mock_aws()

    region = "us-west-2"
    if cloudformation_stack_exists:
        cloud_resource_mock.aws_cloudformation_stack_id = _create_cloudformation_stack(
            region
        )

    result = verify_aws_cloudformation_stack(
        cloud_resource=cloud_resource_mock,
        boto3_session=boto3.Session(region_name=region),
        logger=BlockLogger(),
    )
    assert result == expected_result


@mock_ec2
def test_associate_subnets_with_azs():
    cloud_resource_mock = generate_cloud_resource_mock_aws()
    region = "us-west-2"
    vpc_id = _create_vpc(region=region)
    subnet_ids = _create_subnets(vpc_id=vpc_id, region=region)

    cloud_resource_mock.aws_subnet_ids = subnet_ids
    cloud_resource_mock.aws_subnet_ids_with_availability_zones = None

    new_cloud_resource = associate_subnets_with_azs(cloud_resource_mock, region)
    assert new_cloud_resource.aws_subnet_ids is None
    assert len(new_cloud_resource.aws_subnet_ids_with_availability_zones) == 2



import enum
from typing import Any, Dict, Optional

from pydantic import Field, root_validator

from anyscale.controllers.job_controller import BaseHAJobConfig
from anyscale.utils.runtime_env import override_runtime_env_for_local_working_dir


class UserServiceAccessTypes(str, enum.Enum):
    private = "private"
    public = "public"


class ServiceConfig(BaseHAJobConfig):
    healthcheck_url: Optional[str] = Field(
        None, description="Healthcheck url for service."
    )
    access: UserServiceAccessTypes = Field(
        UserServiceAccessTypes.public,
        description=(
            "Whether user service (eg: serve deployment) can be accessed by public "
            "internet traffic. If public, a user service endpoint can be queried from "
            "the public internet with the provided authentication token. "
            "If private, the user service endpoint can only be queried from within "
            "the same Anyscale cloud and will not require an authentication token."
        ),
    )

    entrypoint: Optional[str] = Field(
        None,
        description="A script that will be run to start your job. This command will be run in the root directory of the specified runtime env. Eg. 'python script.py'",
    )

    ray_serve_config: Optional[Dict[str, Any]] = Field(
        None,
        description=(
            "The Ray Serve config to use for this Production service. It is supported only on v2 clouds."
            "This config defines your Ray Serve application, and will be passed directly to Ray Serve. "
            "You can learn more about Ray Serve config files here: https://docs.ray.io/en/latest/serve/production-guide/config.html"
        ),
    )

    # Version level fields for Services V2
    version: Optional[str] = Field(
        None,
        description="A version string that represents the version for this service. "
        "Will be populated with the hash of the config if not specified.",
    )
    canary_weight: Optional[int] = Field(
        None,
        description="A manual target weight for this service. "
        "If this field is not set, the service will automatically roll out. "
        "If set, this should be a number between 0 and 100 (inclusive). "
        "The newly created version will have weight `canary_weight` "
        "and the existing version will have `100 - canary_weight`.",
    )

    @root_validator
    def validates_config(cls, values) -> Dict[str, Any]:
        # entrypoint is used for Services V1 and serve config for Services V2
        is_entrypoint_present = bool(values.get("entrypoint"))
        is_serve_config_present = bool(values.get("ray_serve_config"))
        assert is_entrypoint_present != is_serve_config_present, (
            "Please specify one of 'entrypoint' or 'ray_serve_config'. "
            "'entrypoint' should be specified on v1 clouds, "
            "and 'ray_serve_config' should be specified on v2 clouds."
        )

        if is_entrypoint_present:
            assert (
                values.get("healthcheck_url") is not None
            ), "healthcheck_url should be set for Services on v1 clouds."
        if is_serve_config_present:
            assert (
                values.get("healthcheck_url") is None
            ), "healthcheck_url should not be set for Services on v2 clouds."
            assert (
                values.get("runtime_env") is None
            ), "runtime_env should not be set for Services on v2 clouds."

        return values

    @root_validator
    def overwrites_runtime_env_in_serve_config(cls, values) -> Dict[str, Any]:
        if values.get("ray_serve_config") and values.get("ray_serve_config").get(
            "runtime_env"
        ):
            runtime_env = values.get("ray_serve_config").get("runtime_env")
            values["ray_serve_config"] = {
                **(values["ray_serve_config"]),
                "runtime_env": override_runtime_env_for_local_working_dir(runtime_env),
            }
        return values



from contextlib import contextmanager
import os
import subprocess
import time
from typing import Any, Dict, Optional, Tuple

import click

# TODO(tchordia) get rid of this once the local run_kill_child is in prod
from anyscale.background.job_runner import (  # pylint:disable=private-import
    _run_kill_child,
)
from anyscale.cli_logger import BlockLogger
from anyscale.client.openapi_client.models.app_config_config_schema import (
    AppConfigConfigSchema,
)
from anyscale.client.openapi_client.models.baseimagesenum import BASEIMAGESENUM
from anyscale.client.openapi_client.models.compute_node_type import ComputeNodeType
from anyscale.client.openapi_client.models.worker_node_type import WorkerNodeType
from anyscale.controllers.base_controller import BaseController
from anyscale.sdk.anyscale_client.models.cluster_compute_config import (
    ClusterComputeConfig,
)
from anyscale.util import (
    DEFAULT_RAY_VERSION,
    get_ray_and_py_version_for_default_cluster_env,
)
from anyscale.utils.cloud_utils import get_default_cloud
from anyscale.utils.imports.all import try_import_ray
from anyscale.utils.ray_version_checker import detect_python_minor_version


def run_command_on_all_nodes(command: str, timeout: Optional[float]):
    class Runner:
        """
        This class is similar to BackgroundJobRunner from job_runner.py.
        The version of run_background_job in BackgroundJobRunner takes an additional BackgroundJobContext argument
        that requires a creator_db_id and sets its own runtime environment.
        This simpler version does not.

        TODO(mattweber): If we ever release this, we will have to creator_db_id, otherwise this won't show in the UI
        """

        def run_background_job(self, command: str) -> None:
            # Update the context with the runtime env uris
            env_vars = {
                "PYTHONUNBUFFERED": "1",  # Make sure python subprocess streams logs https://docs.python.org/3/using/cmdline.html#cmdoption-u
            }
            env = {**os.environ, **env_vars}

            try:
                # TODO(mattweber): Once the publicly named run_kill_child is
                # available on product nodes, remove the underscore on this function.
                _run_kill_child(command, shell=True, check=True, env=env)  # noqa
            finally:
                # allow time for any logs to propogate before the task exits
                time.sleep(1)

    ray = try_import_ray()
    from ray.util.placement_group import placement_group

    nodes = ray.nodes()
    bundles = [{"CPU": 1} for node in nodes]
    pg = placement_group(bundles, strategy="STRICT_SPREAD")
    ray.get(pg.ready(), timeout=timeout)

    runners = [
        ray.remote(Runner)
        .options(  # type: ignore
            placement_group=pg, placement_group_bundle_index=i
        )
        .remote()
        for i, x in enumerate(nodes)
    ]
    futures = [runner.run_background_job.remote(command) for runner in runners]
    ray.get(futures)


class ConfigController(BaseController):
    """
    This controller powers functionalities related to Anyscale configurations such as
    cluster environments for compute configs.
    """

    def __init__(
        self, log: Optional[BlockLogger] = None, initialize_auth_api_client: bool = True
    ):
        if log is None:
            log = BlockLogger()

        super().__init__(initialize_auth_api_client=initialize_auth_api_client)
        self.sdk_client = self.anyscale_api_client

        self.log = log

    def convert_cluster_yaml(
        self,
        cloud_name: Optional[str],
        cluster_yaml: Dict[str, Any],
        ml: bool,
        gpu: bool,
    ) -> Tuple[AppConfigConfigSchema, ClusterComputeConfig]:
        cloud_id, cloud_name = get_default_cloud(self.api_client, cloud_name)
        self.log.info(f"Using cloud: {cloud_name}")

        cluster_env = self._convert_cluster_env(cluster_yaml, ml, gpu)
        compute_config = self._convert_compute_config(cloud_id, cluster_yaml)
        return cluster_env, compute_config

    def create_cluster_env(self, name: str, cluster_env_config: Dict[str, Any]) -> None:
        config = AppConfigConfigSchema(**cluster_env_config)
        self.sdk_client.create_cluster_environment(
            {"name": name, "config_json": config.to_dict()}
        )

    def create_compute_config(
        self, name: str, compute_config: Dict[str, Any], anonymous: bool
    ) -> None:
        cluster_compute_config = ClusterComputeConfig(**compute_config)
        self.sdk_client.create_cluster_compute(
            {"name": name, "config": cluster_compute_config, "anonymous": anonymous}
        )

    def setup_dev_ray(  # noqa: PLR0913
        self,
        working_dir: str,
        cluster_name: str,
        cluster_env: Optional[str],
        cluster_compute: Optional[str],
        timeout: Optional[float],
    ):
        # import ray dependencies here to avoid making ray a general dependency for
        # the CLI
        import ray

        optional_args = {}
        if cluster_env:
            optional_args["cluster_env"] = cluster_env
        if cluster_compute:
            optional_args["cluster_compute"] = cluster_compute

        ray.init(
            f"anyscale://{cluster_name}",
            runtime_env={"working_dir": working_dir},
            **optional_args,
        )
        run_command_on_all_nodes(
            "cp -RT ray /tmp/dev-ray; python /tmp/dev-ray/python/ray/setup-dev.py -y",
            timeout,
        )

    def _convert_node_config_aws(
        self, name: str, node_type_config: Dict[str, Any], is_head_node: bool = False,
    ) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """
        Returns a tuple
        Index 0: a dictionary in the shape of either a ComputeNodeType or a WorkerNodeType depending on
        if `is_head_node` is True.
        Index 1: a dictionary of additional aws options which will be applied globally.
        """
        with self._warn_if_fail("Instance type is required", is_error=True):
            node_config = node_type_config["node_config"]
            instance_type = node_config["InstanceType"]

        min_workers = None
        max_workers = None
        use_spot = None

        if not is_head_node:
            min_workers = node_type_config.get("min_workers")
            max_workers = node_type_config.get("max_workers")
            market_type = node_config.get("InstanceMarketOptions", {}).get("MarketType")
            use_spot = market_type == "spot"

        # No need to handle resources. They will be filled in based on the instance type names

        output_node_config = {
            "name": name,
            "instance_type": instance_type,
            **({"min_workers": min_workers} if min_workers is not None else {}),
            **({"max_workers": max_workers} if max_workers is not None else {}),
            **({"use_spot": use_spot} if use_spot is not None else {}),
        }

        supported_option_keys = [
            "BlockDeviceMappings",
            "IamInstanceProfile",
            "TagSpecifications",
            "NetworkInterfaces",
        ]

        aws_options = {
            key: value
            for key, value in node_config.items()
            if key in supported_option_keys
        }

        return output_node_config, aws_options

    def _convert_compute_config(
        self, cloud_id: str, cluster_yaml: Dict[str, Any]
    ) -> ClusterComputeConfig:
        with self._warn_if_fail("Provider info is required", is_error=True):
            provider_config = cluster_yaml["provider"]
            provider_type = provider_config["type"]
            region = provider_config["region"]

        if provider_type != "aws":
            raise click.ClickException("Only AWS cluster yamls are supported for now.")

        # Provider configs
        availability_zone = provider_config.get("availability_zone")
        allowed_azs = availability_zone.split(",") if availability_zone else None

        # Available node types
        available_node_types = cluster_yaml.get("available_node_types")
        head_node_type = cluster_yaml.get("head_node_type")
        if not available_node_types or not head_node_type:
            # TODO (aguo): Do we need to handle legacy node types anymore?
            raise click.ClickException(
                "Available_node_types and head_node_type fields are required."
            )
        else:
            head_node = None
            worker_nodes = []
            global_aws_options: Dict[str, Any] = {}
            for name, node_type_config in available_node_types.items():
                if name == head_node_type:
                    node_config, aws_options = self._convert_node_config_aws(
                        name, node_type_config, is_head_node=True
                    )
                    head_node = ComputeNodeType(**node_config)
                else:
                    node_config, aws_options = self._convert_node_config_aws(
                        name, node_type_config
                    )
                    worker_nodes.append(WorkerNodeType(**node_config))

                global_aws_options = {**global_aws_options, **aws_options}

            if not head_node:
                raise click.BadArgumentUsage(
                    "available_node_types is missing the node_config for the head_node_type."
                )

            if len(global_aws_options):
                global_aws_options_keys = ", ".join(global_aws_options.keys())
                self.log.warning(
                    f"We squished the following configurations: ({global_aws_options_keys}) "
                    "from each node type config and set them globally across all node types."
                )

                if "IamInstanceProfile" in global_aws_options:
                    self.log.warning(
                        "WARNING: IamInstanceProfile found in cluster yaml. In most cases, "
                        "overriding this is not necessary because the anyscale cloud "
                        "that was selected will set up the correct iams for each node."
                    )

        return ClusterComputeConfig(
            cloud_id=cloud_id,
            max_workers=cluster_yaml.get("max_workers"),
            region=region,
            allowed_azs=allowed_azs,
            head_node_type=head_node,
            worker_node_types=worker_nodes,
            aws=global_aws_options,
        )

    def _convert_cluster_env(
        self, cluster_yaml: Dict[str, Any], ml: bool, gpu: bool,
    ) -> AppConfigConfigSchema:
        self.log.info(
            "Generating cluster environment based on local python and ray versions"
        )

        base_image = self._select_base_image(ml, gpu)

        post_build_commands = []
        setup_commands = cluster_yaml.get("setup_commands", [])
        if len(setup_commands):
            self.log.warning(
                "WARNING: Setup commands detected! These have been converted "
                "to post-build commands for your cluster env.\n\n"
                "IMPORTANT:\n"
                "1. If your setup commands is installing pip packages, "
                "please add those pip packages to the `python` list in your cluster env instead.\n"
                "2. If your setup commands is installing a ray version, make sure to install that "
                "same ray version locally."
            )
            post_build_commands = setup_commands
        head_setup_commands = cluster_yaml.get("head_setup_commands", [])
        worker_setup_commands = cluster_yaml.get("worker_setup_commands", [])
        if len(head_setup_commands) or len(worker_setup_commands):
            self.log.warning(
                "WARNING: Head or worker specific setup commands detected! These will not be "
                "converted in your cluster environment because the same cluster environment will "
                "be installed on the head and worker nodes. Please move these to the post build "
                "commands of the cluster environment to run on all nodes."
            )
        initialization_commands = cluster_yaml.get("initialization_commands", [])
        if len(initialization_commands):
            self.log.warning(
                "WARNING: Initialization commands detected! These will not be converted in your "
                "cluster environment because commands in Anyscale can only be run inside the cluster "
                "environment, whereas initialization commands are run in open source outside "
                "the docker container. Please move these to the post build commands of the "
                "cluster environment to run on all nodes."
            )

        if len(cluster_yaml.get("file_mounts", {})) or len(
            cluster_yaml.get("cluster_synced_files", {})
        ):
            self.log.warning(
                "WARNING: file_mounts or cluster_synced_files detected! These are not converted into a cluster env. "
                "If you need to upload local files into your cluster, please use a runtime environment."
            )

        return AppConfigConfigSchema(
            base_image=base_image,
            post_build_cmds=post_build_commands,
            python={"pip_packages": [], "conda_packages": []},
            debian_packages=[],
            env_vars={},
        )

    def _select_base_image(self, ml: bool, gpu: bool) -> str:
        ray_version, _ = get_ray_and_py_version_for_default_cluster_env()
        python_version = detect_python_minor_version()
        self.log.info(
            f"Detected python version: {python_version}. Detected ray version: {ray_version}"
        )

        if "dev0" in ray_version:
            self.log.warning(
                "There is no built-in base images for nightly versions of Ray. Selecting "
                f"{DEFAULT_RAY_VERSION} as default. If you need to use a newer version of Ray, please install "
                "that Ray version as a post-build command."
            )
            ray_version = DEFAULT_RAY_VERSION

        return str(
            getattr(
                BASEIMAGESENUM,
                self._find_enum_name(python_version, ray_version, ml, gpu),
            )
        )

    def _find_enum_name(
        self, python_version: str, ray_version: str, ml: bool, gpu: bool
    ) -> str:
        enum_name = "ANYSCALE_RAY"
        if ml:
            enum_name += "_ML"

        enum_name += f"_{ray_version.replace('.', '_')}"

        major_version, minor_version = python_version.split(".")
        enum_name += f"_PY{major_version}{minor_version}"

        if gpu:
            enum_name += "_GPU"

        return enum_name

    @contextmanager
    def _warn_if_fail(self, message: str, is_error: bool = False) -> Any:
        try:
            yield
        except KeyError as e:
            if is_error:
                raise click.ClickException(message) from e
            else:
                self.log.warning(message)


# TODO(tchordia) once this function is in prod, remove last dependency from anyscale.background
def run_kill_child(
    *popenargs, input=None, timeout=None, check=False, **kwargs  # noqa: A002
) -> subprocess.CompletedProcess:
    """
    This function is a fork of subprocess.run with fewer args.
    The goal is to create a child subprocess that is GUARANTEED to exit when the parent exits
    This is accomplished by:
    1. Making sure the child is the head of a new process group
    2. Create a third "Killer" process that is responsible for killing the child when the parent dies
    3. Killer process checks every second if the parent is dead.
    4. Killing the entire process group when we want to kill the child

    Arguments are the same as subprocess.run
    """
    # Start new session ensures that this subprocess starts as a new process group
    with subprocess.Popen(*popenargs, start_new_session=True, **kwargs) as process:
        parent_pid = os.getpid()
        child_pid = process.pid
        child_pgid = os.getpgid(child_pid)

        # Open a new subprocess to kill the child process when the parent process dies
        # kill -s 0 parent_pid will succeed if the parent is alive.
        # If it fails, SIGKILL the child process group and exit
        subprocess.Popen(
            f"while kill -s 0 {parent_pid}; do sleep 1; done; kill -9 -{child_pgid}",
            shell=True,
            # Suppress output
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )

        try:
            stdout, stderr = process.communicate(input, timeout=timeout)
        except:  # noqa      (this is taken from subprocess.run directly)
            # Including KeyboardInterrupt, communicate handled that.
            process.kill()
            # We don't call process.wait() as .__exit__ does that for us.
            raise

        retcode = process.poll()
        if check and retcode:
            raise subprocess.CalledProcessError(
                retcode, process.args, output=stdout, stderr=stderr
            )
    return subprocess.CompletedProcess(process.args, retcode or 0, stdout, stderr)



# AUTOGENERATED - modify shared_anyscale_util in root directory to make changes
import dataclasses
from typing import Any, Optional, Type, TypeVar


AWS_ROLE_ARN_RESOURCE_ID_PREFIX = "role/"
_AwsArnT = TypeVar("_AwsArnT", bound="AwsArn")
NoneType = type(None)


@dataclasses.dataclass
class AwsArn:
    partition: str
    service: str
    resource_id: str
    region: Optional[str] = dataclasses.field(default=None)
    account_id: Optional[str] = dataclasses.field(default=None)

    def _validate_fields(self):
        """Enforce typing of each field."""
        if not isinstance(self.partition, str):
            raise ValueError(f"Partition is of type {type(self.partition)}, not str")

        if not isinstance(self.service, str):
            raise ValueError(f"Service is of type {type(self.service)}, not str")

        if not isinstance(self.region, (str, type(None))):
            raise ValueError(
                f"Region is of type {type(self.region)}, not Optional[str]"
            )

        if not isinstance(self.account_id, (str, type(None))):
            raise ValueError(
                "Account ID is of type {}, not Optional[str]".format(
                    type(self.account_id)
                )
            )

        if not isinstance(self.resource_id, str):
            raise ValueError(f"ResourceId is of type {type(self.resource_id)}, not str")

    def _empty_str_to_none(self):
        """Sets empty strings to None."""
        for k, v in dataclasses.asdict(self).items():
            if v == "":
                setattr(self, k, None)

    def __post_init__(self):
        # pydantic.validator("*", pre=True)
        self._empty_str_to_none()

        # Standard pydantic validation
        self._validate_fields()

    @classmethod
    def from_string(cls: Type[_AwsArnT], arn: str) -> _AwsArnT:
        """Convert an AWS ARN string into a Pydantic Model.

        The ARN is a standard format in AWS with
        the following format:
        arn:<partition>:<service>:<region>:<account>:<resource>
        0       1          2         3        4         5

        > arn:aws:apprunner:us-west-2:959243851260:my_app

        Region & Account can be empty i.e:
        > arn:aws:s3:::bucketName
        """
        spl = arn.split(":")
        if len(spl) != 6 or spl[0] != "arn":
            raise ValueError(f"invalid arn: {arn}")
        return cls(
            partition=spl[1],
            service=spl[2],
            region=spl[3],
            account_id=spl[4],
            resource_id=spl[5],
        )

    def to_string(self) -> str:
        return "arn:{p}:{s}:{reg}:{a}:{rid}".format(
            p=self.partition,
            s=self.service,
            reg=self.region or "",
            a=self.account_id or "",
            rid=self.resource_id,
        )


@dataclasses.dataclass
class AwsRoleArn(AwsArn):
    """Role ARNs are of the format arn:aws:iam::<account>:role/<role name>."""

    account_id: str
    region: NoneType

    def _validate_role(self):
        """Validate that a Role has the correct ResourceId."""
        if self.region is not None:
            raise ValueError("Roles do not have a region!")
        if self.account_id is None:
            raise ValueError("No AccountId")
        if not self.resource_id.startswith(AWS_ROLE_ARN_RESOURCE_ID_PREFIX):
            raise ValueError(f"ResourceId {self.resource_id} is not a role")

    def __post_init__(self):
        self.partition = "aws"
        self.service = "iam"

        super().__post_init__()
        self._validate_role()

    @classmethod
    def from_role_name(
        cls: Type[_AwsArnT], account_id: str, role_name: str
    ) -> _AwsArnT:
        return cls(
            partition="aws",
            service="iam",
            account_id=account_id,
            resource_id=f"role/{role_name}",
        )

    def to_role_name(self) -> str:
        """Get the role name from a role ARN.
        If there is a path, this gets the actual name (which is following the last slash).
        Role with path (`/anyscale/`): arn:aws:iam::959243851260:role/anyscale/thomas-test-role
        The Role name *is* `thomas-test-role`
        NOTE: The resource_id for a Role is "role/<role_name>".
        """
        return self.resource_id.split("/")[-1]


def get_dataplane_role_arn(
    account_id: str, cloud_resource_record: Optional[Any],
) -> AwsRoleArn:
    """ Helper method to get instance role from cloud_resource_record. """
    if cloud_resource_record and cloud_resource_record.aws_iam_role_arns:
        return AwsRoleArn.from_string(cloud_resource_record.aws_iam_role_arns[1])

    return AwsRoleArn.from_role_name(
        account_id=account_id, role_name="ray-autoscaler-v1",
    )



import click

from anyscale.controllers.auth_controller import AuthController


@click.group(
    "auth",
    short_help="Configure the Anyscale authentication credentials",
    help="""Configure the Anyscale authentication credentials""",
)
def auth_cli() -> None:
    pass


@auth_cli.command(name="set", help="Set up credentials and save it to a file")
def auth_set() -> None:
    auth_controller = AuthController()
    auth_controller.set()


@auth_cli.command(
    name="show", help="Show the information of the authenticated user using credentials"
)
def auth_show() -> None:
    auth_controller = AuthController()
    auth_controller.show()


@auth_cli.command(
    name="fix", help="Fix the permission of the existing credentials file"
)
def auth_fix() -> None:
    auth_controller = AuthController()
    auth_controller.fix()


@auth_cli.command(name="remove", help="Remove the current credentials file")
def auth_remove() -> None:
    auth_controller = AuthController()
    auth_controller.remove()
